using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using Airplayne.Interface.Util;
using Airplayne.Model;
using Airplayne.Model.Music;
using Airplayne.Model.Repository;
using Airplayne.Properties;
using Airplayne.Tools;
using Airplayne.Tools.Threading;

namespace Airplayne.Interface.Controls
{
  public partial class LibraryControl : UserControl
  {
    private LibraryManager _libraryManager;
    private RequestQueue<string, IEnumerable<Album>> _searchQueue;
    private string _lastRequestedSearch;
    private string _lastCompletedSearch;
    private int _visibleResultCount;

    #region Init
    public LibraryControl()
    {
      InitializeComponent();
      searchBox.ContextMenuStrip = new ContextMenuStrip();
    }

    protected override void OnLoad(EventArgs e)
    {
      base.OnLoad(e);

      if (LibraryManager == null)
        return;

      _visibleResultCount = searchTree.VisibleCount;
      _searchQueue = new RequestQueue<string, IEnumerable<Album>>(Search, 300);
      _searchQueue.ResponseReady += _searchQueue_ResponseReady;
      _searchQueue.ProgressIndication += _searchQueue_ProgressIndication;
      RequestSearch();
      UpdateTop();
    }

    protected override void OnResize(EventArgs e)
    {
      base.OnResize(e);
      _visibleResultCount = searchTree.VisibleCount;
    }

    protected override void OnLayout(LayoutEventArgs e)
    {
      base.OnLayout(e);
      _visibleResultCount = searchTree.VisibleCount;
    }

    void _searchQueue_ProgressIndication(bool working)
    {
      this.TryBeginInvoke(() => progressIndicator.Visible = working);
    }

    void _searchQueue_ResponseReady(string requestParameter, IEnumerable<Album> response)
    {
      if (_lastRequestedSearch == requestParameter)
      {
        _lastCompletedSearch = requestParameter;
        PopulateTree(response);
      }
    }

    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public LibraryManager LibraryManager
    {
      get { return _libraryManager; }
      set
      {
        _libraryManager = value;
        _libraryManager.LibraryUpdated += () => RequestSearch(true);
      }
    }

    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public PlaylistManager PlaylistManager { get; set; }

    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public PlaylistManager BagListManager { get; set; }

    #endregion

    #region Searching
    private void searchBox_TextChanged(object sender, EventArgs e)
    {
      UpdateTop();
      RequestSearch();
    }

    private void UpdateTop()
    {
      bool isRandom = string.IsNullOrEmpty(searchBox.Text); // not strictly true
      string buttonTip = isRandom ? "Reshufle (F5)" : "Clear search (F5)";
      toolTip.SetToolTip(clearRefreshButton, buttonTip);
      clearRefreshButton.Image = isRandom ? Resources.arrow_refresh : Resources.cancel_grey;
      randomInfoLabel.Visible = (string.IsNullOrEmpty(searchBox.Text));
    }

    private void RequestSearch(bool force = false)
    {
      if (_searchQueue == null)
        return;

      string searchString = StringUtils.CleanSearchString(SearchBoxText, false);
      if (force || _lastRequestedSearch != searchString)
      {
        _lastRequestedSearch = searchString;
        _searchQueue.Request(searchString, skipTimeout: force || string.IsNullOrEmpty(searchString));
      }
    }

    private string SearchBoxText
    {
      get
      {
        string text = string.Empty;
        this.TryInvoke(() => { text = searchBox.Text; });
        return text;
      }
    }

    private IEnumerable<Album> Search(string search)
    {
      if (!string.IsNullOrEmpty(search))
      {
        return _libraryManager.Search(search);
      }
      else
      {
        return _libraryManager.RandomAlbums();
      }
    }

    private void PopulateTree(IEnumerable<Album> albums)
    {
      int firstBatchCount = _visibleResultCount + 2;

      TreeNode[] firstBatch = BuildTreeNodes(albums.Take(firstBatchCount).ToList());
      if (!this.TryInvoke(() => AddNodesToTree(firstBatch, true)))
        return;

      TreeNode[] theRest = BuildTreeNodes(albums.Skip(firstBatchCount).ToList());
      this.TryBeginInvoke(() => AddNodesToTree(theRest, false));
    }

    private void AddNodesToTree(TreeNode[] treeNodes, bool clearTree)
    {
      searchTree.BeginUpdate();

      if (clearTree)
        searchTree.Nodes.Clear();

      searchTree.Nodes.AddRange(treeNodes);

      if (clearTree)
        AdjustTreeAfterPopulation();

      searchTree.EndUpdate();
    }

    private void AdjustTreeAfterPopulation()
    {
      if (searchTree.Nodes.Count == 1)
      {
        searchTree.Nodes[0].Expand();
      }
      if (searchTree.Nodes.Count > 0)
      {
        searchTree.Nodes[0].EnsureVisible(); // scroll to top
        searchTree.SelectedNode = searchTree.Nodes[0];
      }
    }

    private static TreeNode[] BuildTreeNodes(IList<Album> albums)
    {
      TreeNode[] nodes = new TreeNode[albums.Count];
      for (int i = 0; i < albums.Count; i++)
      {
        TreeNode aNode = new TreeNode(albums[i].Title);
        aNode.ImageKey = "album";
        aNode.SelectedImageKey = "album";
        aNode.Tag = albums[i];
        aNode.Nodes.Add("dummy");
        nodes[i] = aNode;
      }
      return nodes;
    }

    protected override bool ProcessDialogKey(Keys keyData)
    {
      if (keyData == Keys.Escape)
      {
        ClearRefresh(false);
        return true;
      }
      else if ((keyData & ~Keys.Modifiers) == Keys.Enter)
      {
        Keys mofifiers = keyData & Keys.Modifiers;
        bool shift = keyData.HasFlag(Keys.Shift);
        bool control = keyData.HasFlag(Keys.Control);
        bool alt = keyData.HasFlag(Keys.Alt);

        AddTreeSelectionToPlaylist(asNext: shift, fullAlbum: control, toBag: alt);
        return true;
      }
      else
      {
        return base.ProcessDialogKey(keyData);
      }
    }

    private TrackCollection GetSelectedMusic(bool fullAlbum)
    {
      TreeNode selectedNode = searchTree.SelectedNode;
      if (selectedNode == null)
        return null;

      if (fullAlbum && selectedNode.Parent != null)
        selectedNode = selectedNode.Parent;

      IMusic selected = selectedNode.Tag as IMusic;
      TrackCollection result;

      if (selected is Album)
      {
        result = new TrackCollection((Album)selected, searchBox.Text, !fullAlbum);
      }
      else // if (selected is Track)
      {
        result = new TrackCollection((Track)selected);
      }
      return result;
    }

    private Album SelectedAlbum
    {
      get
      {
        TreeNode selectedNode = searchTree.SelectedNode;
        if (selectedNode == null)
          return null;
        IMusic selected = (selectedNode.Tag as IMusic);
        return selected.AlbumInstance;
      }
    }

    private Track SelectedTrack
    {
      get
      {
        TreeNode selectedNode = searchTree.SelectedNode;
        if (selectedNode == null)
          return null;
        return (selectedNode.Tag as Track);
      }
    }

    private void AddTreeSelectionToPlaylist(bool asNext = false, bool fullAlbum = false, bool toBag = false)
    {
      IMusic music = GetSelectedMusic(fullAlbum);
      if (music == null) return;

      (toBag ? BagListManager : PlaylistManager).InsertTrackRange(music.Tracks, asNext: asNext);
    }

    private void searchTree_NodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e)
    {
      if (searchTree.SelectedNode.Tag is Track)
        PlaylistManager.InsertTrackRange(GetSelectedMusic(false).Tracks);
    }

    private void searchBox_KeyDown(object sender, KeyEventArgs e)
    {
      if (e.Modifiers == Keys.None)
      {
        if (e.KeyCode == Keys.Up || e.KeyCode == Keys.Down)
        {
          HandleVerticalArrow(e);
        }
        else if (e.KeyCode == Keys.Right)
        {
          HandleRightArrow();
        }
        else if (e.KeyCode == Keys.Left)
        {
          HandleLeftArrow();
        }
      }

      if (e.KeyCode == Keys.PageDown)
      {
        HandlePageDown();
      }
      else if (e.KeyCode == Keys.PageUp)
      {
        HandlePageUp();
      }
      else if (e.KeyCode == Keys.Apps || (e.KeyCode == Keys.F10 && e.Modifiers == Keys.Shift))
      {
        HandleAppsKey();
      }
    }

    private void HandleVerticalArrow(KeyEventArgs e)
    {
      if (searchTree.Nodes.Count > 0)
      {
        TreeNode selected = searchTree.SelectedNode;
        if (selected == null)
        {
          searchTree.SelectedNode = searchTree.Nodes[0];
        }
        else
        {
          if (e.KeyCode == Keys.Up)
            searchTree.SelectedNode = selected.PrevVisibleNode ?? selected;
          else
            searchTree.SelectedNode = selected.NextVisibleNode ?? selected;
        }
      }
    }

    private void HandleAppsKey()
    {
      TreeNode selected = searchTree.SelectedNode;
      if (selected != null)
      {
        Point nodeLocation = searchTree.PointToScreen(new Point(selected.Bounds.Left, selected.Bounds.Bottom));
        searchTree.ContextMenuStrip.Show(nodeLocation);
      }
    }

    private void HandleRightArrow()
    {
      TreeNode selected = searchTree.SelectedNode;
      if (selected != null)
      {
        if (!selected.IsExpanded)
          selected.Expand();
        else if (selected.Nodes.Count > 0)
          searchTree.SelectedNode = selected.Nodes[0];
      }
    }

    private void HandleLeftArrow()
    {
      TreeNode selected = searchTree.SelectedNode;
      if (selected != null)
      {
        if (selected.Parent != null)
          searchTree.SelectedNode = selected.Parent;
        else if (searchTree.Nodes.Count > 1)
          selected.Collapse();
      }
    }

    private void HandlePageDown()
    {
      if (searchTree.Nodes.Count > 0)
      {
        if (searchTree.SelectedNode == null)
          searchTree.SelectedNode = searchTree.Nodes[0];

        TreeNode selectMe = searchTree.SelectedNode;
        TreeNode next = searchTree.SelectedNode;
        for (int i = 0; i < searchTree.VisibleCount - 1; i++)
        {
          next = next.NextVisibleNode;
          if (next == null)
            break;
          selectMe = next;
        }
        searchTree.SelectedNode = selectMe;
      }
    }

    private void HandlePageUp()
    {
      if (searchTree.Nodes.Count > 0)
      {
        if (searchTree.SelectedNode == null)
        {
          searchTree.SelectedNode = searchTree.Nodes[0];
          return; // no way up from here
        }

        TreeNode selectMe = searchTree.SelectedNode;
        TreeNode next = searchTree.SelectedNode;
        for (int i = 0; i < searchTree.VisibleCount - 1; i++)
        {
          next = next.PrevVisibleNode;
          if (next == null)
            break;
          selectMe = next;
        }
        searchTree.SelectedNode = selectMe;
      }
    }

    private void randomInfoLabel_Click(object sender, EventArgs e)
    {
      searchBox.Focus();
    }

    private void clearRefreshButton_Click(object sender, EventArgs e)
    {
      ClearRefresh();
    }

    public void ClearRefresh()
    {
      ClearRefresh(true);
    }

    public void ClearRefresh(bool reshuffle)
    {
      searchBox.Focus();
      if (searchBox.Text != string.Empty)
      {
        searchBox.Text = string.Empty;
      }
      else if (reshuffle)
        RequestSearch(true);
    }
    #endregion

    #region Lazy loading tree
    private void searchTree_BeforeExpand(object sender, TreeViewCancelEventArgs e)
    {
      Album a = (Album)e.Node.Tag;
      searchTree.BeginUpdate();
      e.Node.Nodes.Clear();

      foreach (Track t in a.TracksMatching(_lastCompletedSearch))
      {
        TreeNode tNode = new TreeNode(t.Title);
        tNode.ImageKey = "track";
        tNode.SelectedImageKey = "track";
        tNode.Tag = t;
        e.Node.Nodes.Add(tNode);
      }
      searchTree.EndUpdate();
    }

    private void searchTree_AfterCollapse(object sender, TreeViewEventArgs e)
    {
      e.Node.Nodes.Clear();
      e.Node.Nodes.Add("dummy");
    }
    #endregion

    #region Dragging Library -> Playlist
    private void searchTree_ItemDrag(object sender, ItemDragEventArgs e)
    {
      TreeNode draggedNode = ((TreeNode)e.Item);
      searchTree.SelectedNode = draggedNode;

      var tracksDrag = GetDragDataForSelection();
      var dataObject = DragHelper.GetDataObject(tracksDrag);

      this.DoDragDrop(dataObject, DragDropEffects.Link | DragDropEffects.Copy);
    }

    private TracksDragData GetDragDataForSelection()
    {
      var selectedTrack = SelectedTrack;
      var selectedAlbum = SelectedAlbum;
      var trackCollection = GetSelectedMusic(false);

      List<string> paths = null;
      if (selectedTrack == null)
        paths = new List<string>(1) { selectedAlbum.FolderPath };

      var tracksDrag = new TracksDragData(LibraryManager, trackCollection.Tracks, pathsOverride: paths);
      return tracksDrag;
    }
    #endregion

    #region Tree context menu
    private void addToPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AddTreeSelectionToPlaylist();
    }

    private void addToBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AddTreeSelectionToPlaylist(toBag: true);
    }

    private void playNextToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AddTreeSelectionToPlaylist(asNext: true);
    }

    private void addWholeAlbumToPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AddTreeSelectionToPlaylist(fullAlbum: true);
    }

    private void searchTree_MouseDown(object sender, MouseEventArgs e)
    {
      if (e.Button == MouseButtons.Right)
      {
        searchTree.SelectedNode = searchTree.GetNodeAt(e.Location);
      }
    }

    private void treeContextMenu_Opening(object sender, CancelEventArgs e)
    {
      if (searchTree.SelectedNode == null)
      {
        e.Cancel = true;
        return;
      }
      bool trackSelected = (SelectedTrack != null);
      selectRandomAlbumToolStripMenuItem.Available = !trackSelected;
      selectRandomTrackToolStripMenuItem.Available = trackSelected;
    }

    private void exploreFolderToolStripMenuItem_Click(object sender, EventArgs e)
    {
      TrackCollection selection = GetSelectedMusic(false);
      if (selection.Tracks.Count() == 1)
        UIUtils.ShowInExplorer(selection.Tracks.First().FullPath, true);
      else if (selection.Tracks.Count() > 1)
        UIUtils.ShowInExplorer(selection.AlbumInstance.FolderPath, false);
    }

    private void selectRandomAlbumToolStripMenuItem_Click(object sender, EventArgs e)
    {
      searchTree.SelectedNode = searchTree.Nodes[Utils.Random.Next(searchTree.Nodes.Count)];
    }

    private void copyToolStripMenuItem_Click(object sender, EventArgs e)
    {
      var dragData = GetDragDataForSelection();
      DragHelper.Copy(dragData);
    }

    private void selectRandomTrackToolStripMenuItem_Click(object sender, EventArgs e)
    {
      TreeNode expandedAlbum = searchTree.SelectedNode.Parent;
      searchTree.SelectedNode = expandedAlbum.Nodes[Utils.Random.Next(expandedAlbum.Nodes.Count)];
    }

    private void googleAlbumToolStripMenuItem_Click(object sender, EventArgs e)
    {
      UIUtils.GoogleAlbum(SelectedAlbum);
    }

    private void googleTrackToolStripMenuItem_Click(object sender, EventArgs e)
    {
      UIUtils.GoogleTrack(SelectedTrack);
    }

    private void exploreToolStripMenuItem_Click(object sender, EventArgs e)
    {
      if (SelectedTrack != null)
        UIUtils.ShowInExplorer(SelectedTrack.FullPath, true);
      else
        UIUtils.ShowInExplorer(SelectedAlbum.FolderPath, false);
    }

    #endregion

    #region Making friends with the outside world

    public void FocusSearchBox()
    {
      searchBox.Focus();
    }

    public void SelectAlbum(Album album)
    {
      // hack city 2000
      string search = string.Format(@"""{0}""", album.Title);
      _lastRequestedSearch = StringUtils.CleanSearchString(search, false);
      searchBox.Text = search;
      RequestSearch(true);
    }

    #endregion

  }
}