﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Linq;

namespace Airplayne.Interface.Controls
{
  /// <summary>
  /// A slightly extended version of the standard ListView.
  /// - Can draw insertion lines
  /// - Reduces flicker
  /// 
  /// "Manual reordering of items inside a ListView"
  /// http://www.codeproject.com/KB/miscctrl/LVCustomReordering.aspx
  /// </summary>f
  public class ListViewEx : ListView
  {
    // from WinUser.h
    private const int WM_PAINT = 0x000F;

    private int? _lineIndex;
    private bool _lineBelowItem;

    public ListViewEx()
    {
      // Reduce flicker
      SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
    }

    public int ItemHeight
    {
      get
      {
        if (this.Items.Count == 0)
          return 0;

        return Items[0].GetBounds(ItemBoundsPortion.Entire).Height;
      }
    }

    private int IconWidth
    {
      get
      {
        if (this.Items.Count == 0)
          return 0;

        return Items[0].GetBounds(ItemBoundsPortion.Icon).Width;
      }
    }

    public void ClearInsertionLine()
    {
      if (_lineIndex != null)
      {
        _lineIndex = null;
        Invalidate();
      }
    }

    public void SetInsertionLine(int index)
    {
      if (Items.Count == 0)
        return;

      int? lineIndexWas = _lineIndex;
      bool lineWasBelow = _lineBelowItem;

      _lineIndex = Math.Min(index, Items.Count - 1);
      _lineBelowItem = (index >= Items.Count);

      if (lineIndexWas != _lineIndex || lineWasBelow != _lineBelowItem)
        Invalidate();
    }

    [DebuggerStepThrough]
    protected override void WndProc(ref Message m)
    {
      base.WndProc(ref m);

      // We have to take this way (instead of overriding OnPaint()) because the ListView is just a wrapper
      // around the common control ListView and unfortunately does not call the OnPaint overrides.
      if (m.Msg == WM_PAINT)
      {
        if (_lineIndex.HasValue)
        {
          Rectangle rect = Items[_lineIndex.Value].GetBounds(ItemBoundsPortion.Entire);
          DrawInsertionLine(rect.Left + IconWidth, rect.Right,
            _lineBelowItem ? rect.Bottom : rect.Top);
        }
      }
    }

    private void DrawInsertionLine(int X1, int X2, int Y)
    {
      using (Graphics g = this.CreateGraphics())
      {
        Pen pen = new Pen(Brushes.Black);
        g.DrawLine(pen, X1, Y, X2 - 1, Y);
      }
    }

    public List<int> SelectedIndicesList
    {
      get
      {
        return SelectedIndices.Cast<int>().ToList();
      }
    }
  }
}