using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using Airplayne.Interface.Dialogs;
using Airplayne.Interface.Util;
using Airplayne.Model;
using Airplayne.Model.Music;
using Airplayne.Playback;
using Airplayne.Tools;

namespace Airplayne.Interface.Controls
{
  public partial class PlaylistControl : UserControl
  {
    private enum DropTypes { Files, LibraryContent, Reorder, OtherList, Unsupported }
    public delegate void AlbumSelectedHandler(Album album);
    public event AlbumSelectedHandler AlbumSelected;
    private PlaylistManager _playlistManager;

    #region Init
    public PlaylistControl()
    {
      InitializeComponent();
    }

    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public PlaylistManager PlaylistManager
    {
      get { return _playlistManager; }
      set
      {
        _playlistManager = value;
        _playlistManager.ListChanged += _playlistManager_ListChanged;
        _playlistManager.PlayerError += _playlistManager_PlayerError;
        SetContextMenu();
      }
    }

    private void _playlistManager_PlayerError(PlaylistManager.PlayerErrorEventArgs args)
    {
      this.TryInvoke(() =>
      {
        using (PlayerErrorDialog dialog = new PlayerErrorDialog(PlaylistManager.CurrentTrack, args.CanProceed))
        {
          args.ProceedToNextTrack = (dialog.ShowDialog(ParentForm) == DialogResult.OK);
        }
      });
    }

    private void SetContextMenu()
    {
      switch (_playlistManager.ListType)
      {
        case PlaylistManager.ListTypes.Playlist:
          listView.ContextMenuStrip = playlistContext;
          break;
        case PlaylistManager.ListTypes.Bag:
          listView.ContextMenuStrip = bagContext;
          break;
      }
    }

    private void _playlistManager_ListChanged(bool listChanged, bool currentChanged)
    {
      if (currentChanged)
        coverDisplay.LoadCoverFor(_playlistManager.CurrentAlbum);

      this.TryBeginInvoke(() => UpdateListView(listChanged));
    }

    protected override void OnLoad(EventArgs e)
    {
      base.OnLoad(e);

      if (PlaylistManager == null)
        return;

      coverDisplay.LoadCoverFor(PlaylistManager.CurrentAlbum);
      UpdateListView(true);
    }

    #endregion

    #region Reaction to playlist events

    private string GetImageKeyForCurrent(Track currentTrack)
    {
      if (currentTrack != null
        && currentTrack.SearchTitle.Contains("shes not there"/*sic*/)
        && (currentTrack.SearchTitle.Contains("zombies") || currentTrack.AlbumInstance.SearchTitle.Contains("zombies")))
        return "special_dino";
      if (DateTime.Today.Month == 12 && DateTime.Today.Day < 27)
        return "special_reindeer";

      return "current";
    }

    private void UpdateListView(bool listChanged)
    {
      listView.BeginUpdate();

      if (listChanged)
      {
        PopulateListView();
        CloseContextMenu();
      }
      ShowListIndexes();

      listView.EndUpdate();

      if (listChanged)
        AdjustColumns();
    }

    private void CloseContextMenu()
    {
      listView.ContextMenuStrip.Close(ToolStripDropDownCloseReason.CloseCalled);
    }

    private void ShowListIndexes()
    {
      string imageKeyForCurrent = GetImageKeyForCurrent(PlaylistManager.CurrentTrack);

      for (int i = 0; i < listView.Items.Count; i++)
      {
        var item = listView.Items[i];

        item.ImageKey = (i == PlaylistManager.CurrentIndex ? imageKeyForCurrent : "blank");
        item.ForeColor = (i < PlaylistManager.AutopilotIndex ? SystemColors.WindowText : SystemColors.GrayText);
      }
    }

    private void PopulateListView()
    {
      var tracks = PlaylistManager.Tracks;

      // replace differing tracks
      for (int i = 0; i < Math.Min(tracks.Count, listView.Items.Count); i++)
        if (!tracks[i].Equals((Track)listView.Items[i].Tag, true))
          listView.Items[i] = BuildListViewItem(tracks[i]);

      // remove superflous tracks
      for (int i = listView.Items.Count - 1; i >= tracks.Count; i--)
        listView.Items.RemoveAt(i);

      // or add missing tracks
      for (int i = listView.Items.Count; i < tracks.Count; i++)
        listView.Items.Add(BuildListViewItem(tracks[i]));
    }

    private static ListViewItem BuildListViewItem(Track t)
    {
      ListViewItem item = new ListViewItem(new string[] { t.Title, t.AlbumInstance.Title });
      item.Tag = t;
      return item;
    }

    private void AdjustColumns()
    {
      if (PlaylistManager.Tracks.Count > 0)
      {
        trackColumn.AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
        albumColumn.AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize);
      }
      else
      {
        trackColumn.Width = 130;
        albumColumn.Width = 130;
      }
    }

    private void contextMenu_Opening(object sender, CancelEventArgs e)
    {
      if (listView.SelectedIndices.Count == 0)
      {
        e.Cancel = true;
        return;
      }
      bool canShowInLibrary = SelectedTrack.AlbumInstance.AlbumId.HasValue;
      showAlbumInLibraryToolStripMenuItem.Available = canShowInLibrary;
      showInLibraryFromBagToolStripMenuItem.Available = canShowInLibrary;

      acceptAutopilotHereToolStripMenuItem.Available =
        (listView.SelectedIndices[listView.SelectedIndices.Count - 1] >= PlaylistManager.AutopilotIndex);
    }

    #endregion

    #region Interaction

    protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
    {
      if (keyData == (Keys.Control | Keys.A))
        SelectAll();

      return base.ProcessCmdKey(ref msg, keyData);
    }

    private bool RemoveSelectedTracks(string caption = "Remove Track")
    {
      if (!WarnAboutTrackRemoval(caption))
        return false;

      PlaylistManager.RemoveTrackRange(listView.SelectedIndicesList);
      return true;
    }

    private bool WarnAboutTrackRemoval(string caption)
    {
      if (PlaylistManager.PlayerState == Player.States.Playing && PlaylistManager.CurrentIndex.HasValue &&
        listView.SelectedIndicesList.Contains(PlaylistManager.CurrentIndex.Value))
      {
        if (MessageBox.Show("Stop playback and remove the playing track?", caption,
          MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.No)
          return false;
      }
      return true;
    }

    private void listView_ItemActivate(object sender, EventArgs e)
    {
      bool shift = Control.ModifierKeys.HasFlag(Keys.Shift);
      PlaylistManager.ActivateTracks(listView.SelectedIndicesList, shift);
    }

    #endregion

    #region Drag & Drop
    private bool rightDrag = false;
    private const int RIGHTMOUSEBUTTON = 2;
    private List<Track> libraryDropData;
    private int libraryDropIndex;

    private void listView_DragEnter(object sender, DragEventArgs e)
    {
      bool allowed = (e.Data.GetDataPresent(typeof(TracksDragData)) || e.Data.GetDataPresent(DataFormats.FileDrop));
      if (allowed)
      {
        e.Effect = DragDropEffects.Link;
        rightDrag = ((e.KeyState & RIGHTMOUSEBUTTON) > 0);
      }
      else
      {
        e.Effect = DragDropEffects.None;
        rightDrag = false;
      }
    }

    private int GetDropIndex(int x, int y)
    {
      Point targetPoint = listView.PointToClient(new Point(x, y));
      targetPoint.Y += listView.ItemHeight / 2;
      targetPoint.X = 10;
      ListViewItem targetItem = listView.GetItemAt(targetPoint.X, targetPoint.Y);
      int targetIndex = (targetItem == null ? PlaylistManager.Tracks.Count : targetItem.Index);
      return targetIndex;
    }

    private void listView_DragDrop(object sender, DragEventArgs e)
    {
      listView.ClearInsertionLine();

      int targetIndex = GetDropIndex(e.X, e.Y);

      if (e.Data.GetDataPresent(typeof(TracksDragData)))
      {
        HandleTracksDragDataDropped(e, targetIndex);
      }
      else if (e.Data.GetDataPresent(DataFormats.FileDrop))
      {
        HandleFileDrop(e, targetIndex);
      }
    }

    private void HandleTracksDragDataDropped(DragEventArgs e, int targetIndex)
    {
      var tracksDrag = GetTracksDragData(e);
      if (tracksDrag.IsFromLibrary)
      {
        libraryDropData = tracksDrag.Tracks;
        libraryDropIndex = targetIndex;

        if (!rightDrag)
          PlaylistManager.InsertTrackRange(tracksDrag.Tracks, targetIndex);
        else
          rightDragContext.Show(e.X, e.Y);
      }
      else if (tracksDrag.IsFromPlaylist)
      {
        if (tracksDrag.SourceModel == PlaylistManager)
        {
          PlaylistManager.MoveTracks(tracksDrag.Indices, targetIndex);
        }
        else
        {
          var args = new CancelEventArgs();
          tracksDrag.DoCallback(args);
          if (!args.Cancel)
            PlaylistManager.InsertTrackRange(tracksDrag.Tracks, targetIndex);
        }
      }
    }

    private static TracksDragData GetTracksDragData(DragEventArgs e)
    {
      if (e.Data.GetDataPresent(typeof(TracksDragData)))
        return (TracksDragData)e.Data.GetData(typeof(TracksDragData));
      else
        return null;
    }

    private void HandleFileDrop(DragEventArgs e, int targetIndex)
    {
      string[] paths = (string[])e.Data.GetData(DataFormats.FileDrop);
      PlaylistManager.TrackOpener.AddTracks(targetIndex, paths);
    }

    private void insertSelectionToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.InsertTrackRange(libraryDropData, libraryDropIndex);
    }

    private void insertWholeAlbumToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.InsertTrackRange(libraryDropData[0].AlbumInstance.Tracks, libraryDropIndex);
    }

    private void addToPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.MoveTracksToSisterList(listView.SelectedIndicesList);
    }

    private void playNextFromBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.MoveTracksToSisterList(listView.SelectedIndicesList, asNext: true);
    }

    private void moveToBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      if (WarnAboutTrackRemoval("Move To Bag"))
        PlaylistManager.MoveTracksToSisterList(listView.SelectedIndicesList);
    }

    private void listView_ItemDrag(object sender, ItemDragEventArgs e)
    {
      var selectedIndices = listView.SelectedIndicesList;
      var tracksDrag = new TracksDragData(
        PlaylistManager, SelectedTracks,
        indices: selectedIndices,
        callback: (td, args) => { args.Cancel = !RemoveSelectedTracks("Move To Bag"); });

      var dragData = DragHelper.GetDataObject(tracksDrag);

      this.DoDragDrop(dragData, DragDropEffects.Link | DragDropEffects.Copy);
    }

    private void listView_DragOver(object sender, DragEventArgs e)
    {
      if (e.Effect == DragDropEffects.None)
        return;

      int index = GetDropIndex(e.X, e.Y);

      var data = GetTracksDragData(e);

      if (data != null && data.SourceModel != PlaylistManager && index > PlaylistManager.AutopilotIndex)
        index = PlaylistManager.AutopilotIndex;

      listView.SetInsertionLine(index);
    }

    private void listView_DragLeave(object sender, EventArgs e)
    {
      listView.ClearInsertionLine();
    }

    #endregion

    #region Context menu

    private IEnumerable<Track> _SelectedTracks
    {
      get
      {
        return listView.SelectedItems.OfType<ListViewItem>().Select(item => item.Tag).Cast<Track>();
      }
    }

    private Track SelectedTrack // or first if multiple
    {
      get
      {
        return _SelectedTracks.FirstOrDefault();
      }
    }

    private List<Track> SelectedTracks
    {
      get
      {
        return _SelectedTracks.ToList();
      }
    }

    private void exploreFolderToolStripMenuItem_Click(object sender, EventArgs e)
    {
      Track t = SelectedTrack;
      if (t != null)
        UIUtils.ShowInExplorer(t.FullPath, true);
    }

    private void searchGoogleForAlbumToolStripMenuItem_Click(object sender, EventArgs e)
    {
      UIUtils.GoogleAlbum(SelectedTrack);
    }

    private void searchGoogleForTrackToolStripMenuItem_Click(object sender, EventArgs e)
    {
      UIUtils.GoogleTrack(SelectedTrack);
    }

    private void moveUpToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.MoveTracks(listView.SelectedIndicesList, listView.SelectedIndices[0] - 1);
    }

    private void moveDownToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.MoveTracks(listView.SelectedIndicesList,
        listView.SelectedIndices[listView.SelectedIndices.Count - 1] + 2);
    }

    private void playNextToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.MoveTracks(listView.SelectedIndicesList, asNext: true);
    }

    private void removeToolStripMenuItem_Click(object sender, EventArgs e)
    {
      RemoveSelectedTracks();
    }

    private void showAlbumInLibraryToolStripMenuItem_Click(object sender, EventArgs e)
    {
      var handler = AlbumSelected;
      if (handler != null)
        handler(SelectedTrack.AlbumInstance);
    }

    private void acceptAutopilotHereToolStripMenuItem_Click(object sender, EventArgs e)
    {
      PlaylistManager.AutopilotIndex = listView.SelectedIndices[listView.SelectedIndices.Count - 1] + 1;
    }

    private void copyToolStripMenuItem_Click(object sender, EventArgs e)
    {
      var tracksDrag = new TracksDragData(PlaylistManager, SelectedTracks);
      DragHelper.Copy(tracksDrag);
    }

    #endregion

    #region Cover
    private void coverDisplay_MouseEnter(object sender, EventArgs e)
    {
      coverDisplay.HideCover();
    }

    private void coverDisplay_DragEnter(object sender, DragEventArgs e)
    {
      coverDisplay.HideCover();
    }

    private void listView_MouseMove(object sender, MouseEventArgs e)
    {
      if (!coverDisplay.RectangleToScreen(coverDisplay.ClientRectangle).Contains(listView.PointToScreen(e.Location)))
        coverDisplay.ShowCover();
    }

    private void listView_MouseLeave(object sender, EventArgs e)
    {
      coverDisplay.ShowCover();
    }
    #endregion

    #region Playlist search

    public int? SearchPlaylist(string search)
    {
      if (!Visible)
        return null;

      string fullSearchString, cleanSearch;
      string[] terms = StringUtils.MakeSearchTerms(search, out fullSearchString, out cleanSearch);
      if (terms.Length == 0)
      {
        DeSelectAll();
        return null;
      }

      return HighlightMatches(terms);
    }

    private int HighlightMatches(string[] terms)
    {
      int count = 0;
      for (int i = 0; i < PlaylistManager.Tracks.Count; i++)
      {
        bool matches = PlaylistManager.Tracks[i].MatchesInAlbum(terms);
        listView.Items[i].Selected = matches;
        if (matches)
        {
          if (count == 0)
          {
            listView.FocusedItem = listView.Items[i];
            listView.EnsureVisible(i);
          }
          count++;
        }
      }
      return count;
    }

    private void DeSelectAll()
    {
      foreach (int index in listView.SelectedIndices)
        listView.Items[index].Selected = false;
    }

    private void SelectAll()
    {
      foreach (ListViewItem item in listView.Items)
        item.Selected = true;
    }

    #endregion
  }
}
