using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.InteropServices;
using System.Security.Principal;
using System.Windows.Forms;
using Airplayne.Interface.Util;
using Airplayne.Model;
using Airplayne.Playback;
using Airplayne.Interface.Hotkeys;

namespace Airplayne.Interface.Dialogs
{
  public partial class OptionsDialog : Form
  {
    private SettingsManager _settingsManager;

    public OptionsDialog(SettingsManager settingsManager)
    {
      UIUtils.SetFont(this);

      InitializeComponent();

      _settingsManager = settingsManager;
    }

    protected override void OnLoad(EventArgs e)
    {
      LoadSettings();

      base.OnLoad(e);
    }

    protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
    {
      if (DialogResult == DialogResult.OK)
      {
        SaveAllSettings();
      }
      base.OnClosing(e);
    }

    private void LoadSettings()
    {
      LoadPlayerSettings();
      LoadHotkeySettings();

      SelectedFolders = _settingsManager.Folders;
      Extensions = _settingsManager.Extensions;
    }

    #region Hotkeys

    private void LoadHotkeySettings()
    {
      foreach (var assignment in _settingsManager.Hotkeys)
        AddHotkeyAssignment(assignment, false);
      AdjustHotkeyListColumns();
    }

    private void AddHotkeyAssignment(HotkeyAssignment assignment, bool adjustColumns = true)
    {
      hotkeyListView.Items.Add(
        new ListViewItem(
          new string[] { 
              assignment.HotkeyAction.Description(),
              assignment.Hotkey.ToString() 
            })
        {
          Tag = assignment
        });
      if (adjustColumns)
        AdjustHotkeyListColumns();
    }

    private void AdjustHotkeyListColumns()
    {
      hotkeyListView.AutoResizeColumns(ColumnHeaderAutoResizeStyle.ColumnContent);
    }

    private void addHotkeyButton_Click(object sender, EventArgs e)
    {
      var assignment = HotkeyAssignmentDialog.SelectHotkeyAssignment();
      if (assignment != null)
        AddHotkeyAssignment(assignment);
    }

    private void removeHotkeyButton_Click(object sender, EventArgs e)
    {
      RemoveSelectedHotkey();
    }

    private void RemoveSelectedHotkey()
    {
      if (hotkeyListView.SelectedIndices.Count > 0)
        hotkeyListView.Items.RemoveAt(hotkeyListView.SelectedIndices[0]);
    }

    private void hotkeyListView_KeyUp(object sender, KeyEventArgs e)
    {
      if (e.KeyCode == Keys.Delete)
        RemoveSelectedHotkey();
    }

    private void SaveHotkeySettings()
    {
      _settingsManager.Hotkeys =
        hotkeyListView.Items.Cast<ListViewItem>()
        .Select(item => item.Tag)
        .Cast<HotkeyAssignment>()
        .ToList();
    }

    #endregion

    private void LoadPlayerSettings()
    {
      var players = _settingsManager.AvailablePlayersByName.ToArray();
      currentPlayerDrop.Items.AddRange(players);
      currentPlayerDrop.SelectedItem = _settingsManager.PlayerType;
    }

    private void SaveAllSettings()
    {
      _settingsManager.Folders = SelectedFolders;
      _settingsManager.Extensions = Extensions;

      SaveHotkeySettings();
    }

    private List<string> SelectedFolders
    {
      get
      {
        List<string> res = new List<string>(folderList.Items.Count);
        foreach (object folder in folderList.Items)
        {
          res.Add(folder.ToString());
        }
        return res;
      }
      set
      {
        folderList.Items.Clear();
        foreach (string folder in value)
          folderList.Items.Add(folder);
      }
    }

    private List<string> Extensions
    {
      get
      {
        string[] arr = extensionsBox.Text.Trim().Split(' ');
        List<string> res = new List<string>(arr.Length);
        foreach (string str in arr)
          res.Add(str.Trim());
        return res;
      }
      set
      {
        extensionsBox.Text = string.Join(" ", value.ToArray());
      }
    }

    public static bool ShowOptionsDialog(SettingsManager settingsManager)
    {
      using (OptionsDialog ls = new OptionsDialog(settingsManager))
      {
        return (ls.ShowDialog() == DialogResult.OK);
      }
    }

    private void removeButton_Click(object sender, EventArgs e)
    {
      int wasSelected = folderList.SelectedIndex;
      if (wasSelected >= 0)
      {
        folderList.Items.RemoveAt(folderList.SelectedIndex);
        if (folderList.Items.Count > wasSelected)
          folderList.SelectedIndex = wasSelected;
      }
    }

    private void folderList_KeyUp(object sender, KeyEventArgs e)
    {
      if (e.KeyCode == Keys.Delete)
        removeButton_Click(sender, e);
    }

    private void addFolderButton_Click(object sender, EventArgs e)
    {
      using (AddFolderDialog dlg = new AddFolderDialog())
      {
        if (dlg.ShowDialog(this) == DialogResult.OK)
        {
          folderList.Items.Add(dlg.Folder);
        }
      }
    }

    private void currentPlayerDrop_SelectionChangeCommitted(object sender, EventArgs e)
    {
      okButton.Enabled = cancelButton.Enabled = currentPlayerDrop.Enabled = false;
      playerSwitchWorker.RunWorkerAsync((SettingsManager.PlayerDescription)currentPlayerDrop.SelectedItem);
    }

    private void playerSwitchWorker_DoWork(object sender, DoWorkEventArgs e)
    {
      _settingsManager.PlayerType = (SettingsManager.PlayerDescription)e.Argument;
    }

    private void playerSwitchWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
    {
      if (e.Error != null)
      {
        if (e.Error is PlayerInitializationException)
        {
          currentPlayerDrop.SelectedItem = _settingsManager.PlayerType;
          MessageBox.Show(this, "Verify that the selected player is properly installed and try again." + Environment.NewLine +
          "See 'Playback engines' in help window.",
            "Failed setting player", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
        else
          throw new ApplicationException("Unexpected exception while switching player", e.Error);
      }

      okButton.Enabled = cancelButton.Enabled = currentPlayerDrop.Enabled = true;
    }

    #region Shield magic
    // http://www.codeproject.com/KB/vista-security/UAC_Shield_for_Elevation.aspx

    [DllImport("user32")]
    public static extern UInt32 SendMessage
        (IntPtr hWnd, UInt32 msg, UInt32 wParam, UInt32 lParam);

    internal const int BCM_FIRST = 0x1600; //Normal button

    internal const int BCM_SETSHIELD = (BCM_FIRST + 0x000C); //Elevated button

    static internal bool IsAdmin()
    {
      WindowsIdentity id = WindowsIdentity.GetCurrent();
      WindowsPrincipal p = new WindowsPrincipal(id);
      return p.IsInRole(WindowsBuiltInRole.Administrator);
    }

    static internal void AddShieldToButton(Button b)
    {
      b.FlatStyle = FlatStyle.System;
      SendMessage(b.Handle, BCM_SETSHIELD, 0, 0xFFFFFFFF);
    }
    #endregion

  }
}