using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using System.Diagnostics;

namespace Airplayne.Interface.Hotkeys
{
  // http://www.liensberger.it/web/blog/?p=207
  public sealed class KeyboardHook : IDisposable
  {
    [DllImport("user32.dll")]
    private static extern bool RegisterHotKey(IntPtr hWnd, int id, uint fsModifiers, uint vk);
    [DllImport("user32.dll")]
    private static extern bool UnregisterHotKey(IntPtr hWnd, int id);

    /// <summary>
    /// Represents the window that is used internally to get the messages.
    /// </summary>
    private class Window : NativeWindow, IDisposable
    {
      private const int WM_HOTKEY = 0x0312;

      public Window()
      {
        this.CreateHandle(new CreateParams() { });
      }

      /// <summary>
      /// Overridden to get the notifications.
      /// </summary>
      /// <param name="m"></param>
      [DebuggerStepThrough]
      protected override void WndProc(ref Message m)
      {
        base.WndProc(ref m);

        switch (m.Msg)
        {
          case WM_HOTKEY:
            HandleHotkey(m);
            break;
        }
      }

      private void HandleHotkey(Message m)
      {
        Keys key = (Keys)(((int)m.LParam >> 16) & 0xFFFF);
        Hotkey.Modifiers modifier = (Hotkey.Modifiers)((int)m.LParam & 0xFFFF);

        if (KeyPressed != null)
          KeyPressed(new Hotkey(modifier, key));
      }

      public event KeyPressedHandler KeyPressed;

      public void Dispose()
      {
        this.DestroyHandle();
      }
    }

    private Window _window = new Window();
    private int _currentId;

    public KeyboardHook()
    {
      // register the event of the inner native window.
      _window.KeyPressed += (hotkey) => OnKeyPressed(hotkey);
    }

    private void OnKeyPressed(Hotkey hotkey)
    {
      var handler = KeyPressed;
      if (handler != null)
        handler(hotkey);
    }

    /// <summary>
    /// Registers a hot key in the system.
    /// </summary>
    /// <param name="modifier">The modifiers that are associated with the hot key.</param>
    /// <param name="key">The key itself that is associated with the hot key.</param>
    public bool RegisterHotKey(Hotkey hotkey)
    {
      _currentId++;

      // register the hot key.
      return RegisterHotKey(_window.Handle, _currentId, (uint)hotkey.Modifier, (uint)hotkey.Key);
    }

    /// <summary>
    /// A hot key has been pressed.
    /// </summary>
    public event KeyPressedHandler KeyPressed;
    public delegate void KeyPressedHandler(Hotkey hotkey);

    public void Dispose()
    {
      UnregisterAllHotkeys();

      // dispose the inner native window.
      _window.Dispose();
    }

    public void UnregisterAllHotkeys()
    {
      for (int i = _currentId; i > 0; i--)
        UnregisterHotKey(_window.Handle, i);
      _currentId = 0;
    }

  }
}
