using System;
using System.ComponentModel;
using System.Windows.Forms;
using Airplayne.Interface.Controls;
using Airplayne.Interface.Dialogs;
using Airplayne.Interface.Util;
using Airplayne.Model;
using Airplayne.Model.Music;
using Airplayne.Model.Repository;
using Airplayne.Playback;
using Microsoft.Win32;
using System.Diagnostics;

namespace Airplayne.Interface
{
  public partial class MainForm : Form
  {
    #region Privates
    private SettingsManager _settingsManager;
    private PlaylistManager _playlistManager;
    private BagListManager _bagListManager;
    private LibraryManager _libraryManager;
    private OpenSaveHelper _openSaveHelper;
    private Timer _statusStripHideTimer = new Timer()
    {
      Interval = 30000,
      Enabled = false
    };
    private bool _sessionIsEnding = false;
    #endregion

    #region Init
    public MainForm(PlaylistManager playlistManager, BagListManager bagListManager, SettingsManager settingsManager,
      LibraryManager libraryManager)
    {
      UIUtils.SetFont(this);

      InitializeComponent();

      _playlistManager = playlistManager;
      _bagListManager = bagListManager;
      _settingsManager = settingsManager;
      _libraryManager = libraryManager;
      _openSaveHelper = new OpenSaveHelper(_settingsManager);

      musicView.PlaylistManager = _playlistManager;
      musicView.BagListManager = _bagListManager;
      musicView.LibraryManager = _libraryManager;
      musicView.SettingsManager = _settingsManager;

      Scanner.ScanProgress += Scanner_ScanProgress;

      _playlistManager.ListChanged += Playlist_TrackChanged;
      _playlistManager.PlayerStateChanged += Player_StateChanged;
      _playlistManager.StopAfterCurrentChanged += ShowStopAfterCurrentState;
      _playlistManager.AutopilotSettingChanged += () => { ShowAutopilotMode(); };
      _settingsManager.VolumeChanged += _settingsManager_VolumeChanged;
      _playlistManager.PlayerPositionChanged += Player_PositionChanged;
      _statusStripHideTimer.Tick += statusStripHideTimer_Tick;
      timeLabel.TimeSelected += timeLabel_TimeSelected;
      volumeLabel.ValueSelected += volumeLabel_ValueSelected;
    }

    protected override void OnLoad(EventArgs e)
    {
      base.OnLoad(e);

      ShowAutopilotMode();
      UpdateInfo();
      volumeLabel.ShowVolume(_settingsManager.Volume);
    }

    void volumeLabel_ValueSelected(int value)
    {
      _settingsManager.Volume = value;
    }

    void timeLabel_TimeSelected(TimeSpan time)
    {
      _playlistManager.SetPlayerPosition(time);
    }

    void Player_PositionChanged(TimeSpan newPosition, TimeSpan total)
    {
      this.TryBeginInvoke(() => timeLabel.ShowPosition(newPosition, total));
    }

    void _settingsManager_VolumeChanged(int volume)
    {
      this.TryBeginInvoke(() => volumeLabel.ShowVolume(volume));
    }

    void Player_StateChanged(Player.States newState)
    {
      this.TryBeginInvoke(() =>
      {
        if (newState != Player.States.Loading)
        {
          playTool.Image = newState == Player.States.Playing
          ? Properties.Resources.control_pause_blue
          : Properties.Resources.control_play_blue;
          playTool.Text = newState == Player.States.Playing ?
            "Pause (Alt+C)" : "Play (Alt+X)";

          if (newState == Player.States.Stopped)
            timeLabel.Empty = true;
        }
        UpdateInfo();
      });
    }

    private void ShowStopAfterCurrentState()
    {
      this.TryBeginInvoke(() =>
      {
        stopSplit.Image = _playlistManager.StopAfterCurrent ?
          Properties.Resources.control_stop_accept_blue :
          Properties.Resources.control_stop_blue;
        stopAfterCurrentToolStripMenuItem.Checked = _playlistManager.StopAfterCurrent;
      });
    }

    private void ShowAutopilotMode()
    {
      this.TryBeginInvoke(() =>
      {
        switch (_playlistManager.AutopilotSetting)
        {
          case Playlist.AutopilotMode.Disabled:
            disabledToolStripMenuItem.Checked = true;
            addAlbumsToolStripMenuItem.Checked = false;
            addTracksToolStripMenuItem.Checked = false;
            autopilotDropMenu.Image = Properties.Resources.cross;
            break;
          case Playlist.AutopilotMode.AddAlbums:
            addAlbumsToolStripMenuItem.Checked = true;
            disabledToolStripMenuItem.Checked = false;
            addTracksToolStripMenuItem.Checked = false;
            autopilotDropMenu.Image = Properties.Resources.cd;
            break;
          case Playlist.AutopilotMode.AddTracks:
            this.addTracksToolStripMenuItem.Checked = true;
            addAlbumsToolStripMenuItem.Checked = false;
            disabledToolStripMenuItem.Checked = false;
            autopilotDropMenu.Image = Properties.Resources.music;
            break;
        }
      });
    }

    public void Scanner_ScanProgress(bool done, int albums, int tracks, int? percent)
    {
      this.TryBeginInvoke(() =>
      {
        ShowStatusStrip();
        UpdateProgressBar(done, percent);

        string count = string.Format("{0:#,0} track{2} in {1:#,0} album{3}", tracks, albums,
          (tracks != 1 ? "s" : string.Empty), (albums != 1 ? "s" : string.Empty));

        if (done)
        {
          statusLabel.Text = count;
          HideStatusStrip();
        }
        else
        {
          statusLabel.Text = string.Format("Scanning, found {0} so far...", count);
        }
      });
    }

    private void UpdateProgressBar(bool done, int? percent)
    {
      if (percent.HasValue)
      {
        statusProgress.Value = percent.Value;
        statusProgress.Style = ProgressBarStyle.Blocks;
      }
      else
        statusProgress.Style = ProgressBarStyle.Marquee;

      statusProgress.Visible = !done;
    }

    private void ShowStatusStrip()
    {
      _statusStripHideTimer.Stop();
      statusStrip.Visible = true;
    }

    private void HideStatusStrip()
    {
      _statusStripHideTimer.Start();
    }

    private void statusStripHideTimer_Tick(object sender, EventArgs e)
    {
      statusStrip.Visible = false;
    }
    #endregion

    #region Track info

    public void Playlist_TrackChanged(bool listChanged, bool currentChanged)
    {
      if (currentChanged)
        this.TryBeginInvoke(() => UpdateInfo());
    }

    private void UpdateInfo()
    {
      Player.States state = _playlistManager.PlayerState;
      Track track = _playlistManager.CurrentTrack;

      ShowStateText(state, track);
      ShowStateIcon(state);
    }

    private void ShowStateText(Player.States state, Track track)
    {
      string info = string.Empty;
      string infoWithState = string.Empty;
      if (track != null)
      {
        info = string.Format("{0} [{1}]", track.Title, track.AlbumInstance.Title);
        infoWithState = state.ToString() + ": " + info;
      }
      else
        infoWithState = state.ToString();

      if (track == null || state == Player.States.Stopped)
        this.Text = "Airplayne";
      else if (state == Player.States.Playing)
        this.Text = info;
      else
        this.Text = infoWithState;
    }

    private void ShowStateIcon(Player.States state)
    {
      switch (state)
      {
        case Player.States.Loading:
          stateLabel.Image = Properties.Resources.state_loading;
          break;
        case Player.States.Paused:
          stateLabel.Image = Properties.Resources.state_paused;
          break;
        case Player.States.Playing:
          stateLabel.Image = Properties.Resources.state_playing;
          break;
        case Player.States.Stopped:
          stateLabel.Image = Properties.Resources.state_stopped;
          break;
      }
      stateLabel.ToolTipText = state.ToString();
    }
    #endregion

    #region Play control

    private void prevButton_Click(object sender, EventArgs e)
    { _playlistManager.Prev(); }
    private void playButton_Click(object sender, EventArgs e)
    { _playlistManager.Play(); }
    private void pauseButton_Click(object sender, EventArgs e)
    { _playlistManager.PlayPause(); }
    private void stopButton_Click(object sender, EventArgs e)
    { _playlistManager.Stop(); }
    private void nextButton_Click(object sender, EventArgs e)
    { _playlistManager.Next(); }

    #endregion

    #region Menu
    private void searchInPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      musicView.PlaylistSearch();
    }

    private void searchBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      musicView.PlaylistSearch(bag: true);
    }

    private void selectFoldersAndFiletypesToolStripMenuItem_Click(object sender, EventArgs e)
    {
      OptionsDialog.ShowOptionsDialog(_settingsManager);
    }

    private void scanInBackgroundToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _libraryManager.ScanInBackground();
    }

    private void exitToolStripMenuItem_Click(object sender, EventArgs e)
    {
      this.Close();
    }

    private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
    {
      HelpDialog.ShowHelp();
    }

    private void disabledToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.AutopilotSetting = Playlist.AutopilotMode.Disabled;
    }

    private void addAlbumsToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.AutopilotSetting = Playlist.AutopilotMode.AddAlbums;
    }

    private void addTracksToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.AutopilotSetting = Playlist.AutopilotMode.AddTracks;
    }

    private void stopAfterCurrentToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.StopAfterCurrent = stopAfterCurrentToolStripMenuItem.Checked;
    }

    private void savePlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _openSaveHelper.SavePlaylist(_playlistManager);
    }

    private void shufflePlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.RandomizePlaylist();
    }

    private void clearPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      ClearList(_playlistManager, "current playlist");
    }

    private void ClearList(PlaylistManager listManager, string name)
    {
      if (MessageBox.Show(string.Format("Remove all tracks from {0}?", name),
        string.Format("Clear {0}", listManager.ListType.Description()), MessageBoxButtons.YesNo, MessageBoxIcon.Question,
        MessageBoxDefaultButton.Button2) == DialogResult.Yes)
      {
        listManager.Clear();
      }
    }

    private void clearBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      ClearList(_bagListManager, "the bag");
    }

    private void reversePlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.ReversePlaylist();
    }

    private void inOrderToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _playlistManager.OrderPlaylist();
    }

    private void viewDebugLogToolStripMenuItem_Click(object sender, EventArgs e)
    {
      using (LogWindow dialog = new LogWindow())
      {
        dialog.ShowDialog(this);
      }
    }

    private void addFilesToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _openSaveHelper.AddTracks(_playlistManager);
    }

    private void openPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      if (_playlistManager.PlayerState != Player.States.Playing ||
        MessageBox.Show(this, "Opening a playlist now will stop playback.", "Open Playlist",
        MessageBoxButtons.OKCancel, MessageBoxIcon.Information, MessageBoxDefaultButton.Button2) == DialogResult.OK)
        _openSaveHelper.OpenPlaylist(_playlistManager);
    }

    private void showBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      showBagToolStripMenuItem.Checked = musicView.ToggleBag();
    }

    private void saveasPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _openSaveHelper.SavePlaylist(_bagListManager);
    }

    private void openFromPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _openSaveHelper.OpenPlaylist(_bagListManager);
    }

    private void addFilesToBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _openSaveHelper.AddTracks(_bagListManager);
    }

    private void randomizeBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _bagListManager.RandomizePlaylist();
    }

    private void reverseBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _bagListManager.ReversePlaylist();
    }

    private void inOrderBagToolStripMenuItem_Click(object sender, EventArgs e)
    {
      _bagListManager.OrderPlaylist();
    }
    #endregion

    #region Form events
    protected override void OnClosing(CancelEventArgs e)
    {
      if (!ConfirmClose())
        e.Cancel = true;
      if (!e.Cancel)
        Program.OnShutdown();

      base.OnClosing(e);
    }

    /// <summary>
    /// </summary>
    /// <returns>True if confirmed</returns>
    private bool ConfirmClose()
    {
      if (!_sessionIsEnding)
      {
        if (_playlistManager.PlayerState == Player.States.Playing)
        {
          DialogResult res = MessageBox.Show(this, "Stop playback and close Airplayne?", "Close Airplayne",
            MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2);
          if (res != DialogResult.Yes)
            return false;
        }
      }
      _sessionIsEnding = false;
      return true;
    }

    private void MainForm_KeyDown(object sender, KeyEventArgs e)
    {
      if (e.Modifiers == Keys.Alt)
      {
        switch (e.KeyCode)
        {
          case Keys.Up:
            _settingsManager.Volume++;
            break;
          case Keys.Down:
            _settingsManager.Volume--;
            break;
        }
      }
      musicView.HandleKey(e);
    }

    protected override bool ProcessMnemonic(char charCode)
    {
      switch (charCode)
      {
        case 'z':
          _playlistManager.Prev();
          return true;
        case 'x':
          _playlistManager.Play();
          return true;
        case 'c':
          _playlistManager.PlayPause();
          return true;
        case 'v':
          _playlistManager.Stop();
          return true;
        case 'b':
          _playlistManager.Next();
          return true;
        case 'f':
          libraryDropMenu.ShowDropDown();
          return true;
      }
      return base.ProcessMnemonic(charCode);
    }

    private void MainForm_Shown(object sender, EventArgs e)
    {
      if (_settingsManager.Folders.Count == 0)
      {
        if (MessageBox.Show("Would you like to add music to the library now?", "Welcome!",
          MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
        {
          OptionsDialog.ShowOptionsDialog(_settingsManager);
        }
      }
    }
    #endregion

    #region WndProc ftl

    [DebuggerStepThrough]
    protected override void WndProc(ref Message m)
    {
      if (m.Msg == 0x11) // WM_QUERYENDSESSION
        _sessionIsEnding = true;

      base.WndProc(ref m);
    }

    #endregion

  }
}