using System;
using System.Collections.Generic;
using Airplayne.Tools;
using System.Linq;

namespace Airplayne.Model.Music
{
  public class Album : IComparable, IMusic
  {
    private List<string> searchWords;

    public Album(string folderPath, string title, List<Track> tracks)
    {
      this.FolderPath = folderPath;
      this.Tracks = tracks;
      this.Title = title;
      this.SearchTitle = StringUtils.CleanSearchString(title, true);
      this.searchWords = StringUtils.GetWords(SearchTitle);
      foreach (Track t in tracks)
      {
        t.AlbumInstance = this;
      }
    }

    #region Properties
    
    public string SearchTitle
    {
      get;
      private set;
    }

    public List<Track> Tracks
    {
      get;
      private set;
    }

    public Album AlbumInstance
    {
      get { return this; }
    }

    public string FolderPath
    {
      get;
      private set;
    }

    public string Title
    {
      get;
      private set;
    }

    public int? AlbumId { get; set; }

    #endregion

    #region Matching, scoring and such

    const int BASE_MATCH_SCORE = 1; // once per album
    const int ALL_TERMS_IN_ALBUM_TITLE = 1000000; // once per album
    const int ALBUM_TITLE_CONTAINS_FULL_SEARCH_STRING = 1000000; // once per album
    const int ALBUM_WORD_SCORE = 100000; // 1-2 per album (once for startwith, once for equals)
    const int ALBUM_TITLE_STARTS_WITH_TERM = 10000; //once per album

    const int TRACK_TITLE_CONTAINS_FULL_SEARCH_STRING = 100;
    const int TRACK_WORD_SCORE = 1;

    public int MatchScore(string[] terms, string fullString)
    {
      List<string> termsNotInAlbumTitle;

      if (!Matches(terms, out termsNotInAlbumTitle))
      {
        return 0;
      }

      int score = BASE_MATCH_SCORE;
      if (termsNotInAlbumTitle.Count == 0)
      {
        score += GetAlbumTitleScore(terms, fullString);
      }
      else
      {
        score += GetTrackScore(termsNotInAlbumTitle, fullString);
      }

      return score;
    }

    private int GetTrackScore(List<string> termsNotInAlbumTitle, string fullString)
    {
      int score = 0;
      // you're usually interested in either an album or a single track, so we'll just look at one track
      // for performance.
      Track first = FirstTrackMatching(termsNotInAlbumTitle);

      if (first.SearchTitle.Contains(fullString))
        score += TRACK_TITLE_CONTAINS_FULL_SEARCH_STRING;

      score += GetWordScore(first.SearchWords, termsNotInAlbumTitle, TRACK_WORD_SCORE);
      return score;
    }

    private int GetAlbumTitleScore(string[] terms, string fullString)
    {
      int score = ALL_TERMS_IN_ALBUM_TITLE;

      if (SearchTitle.Contains(fullString))
        score += ALBUM_TITLE_CONTAINS_FULL_SEARCH_STRING;

      foreach (string term in terms)
      {
        if (SearchTitle.StartsWith(term))
        {
          score += ALBUM_TITLE_STARTS_WITH_TERM;
          break;
        }
      }

      score += GetWordScore(this.searchWords, terms, ALBUM_WORD_SCORE);

      return score;
    }

    public bool Matches(string[] terms)
    {
      List<string> dummy = null;
      return Matches(terms, out dummy);
    }

    private bool Matches(string[] terms,
      out List<string> termsNotInAlbumTitle)
    {
      termsNotInAlbumTitle = GetTermsNotInAlbumTitle(terms);
      return AnyTracksMatch(termsNotInAlbumTitle);
    }

    private List<string> GetTermsNotInAlbumTitle(IEnumerable<string> terms)
    {
      var termsNotInAlbumTitle = new List<string>();

      foreach (string term in terms)
      {
        if (!SearchTitle.Contains(term))
          termsNotInAlbumTitle.Add(term);
      }

      return termsNotInAlbumTitle;
    }

    public List<Track> TracksMatching(string search)
    {
      string fullSearchString;
      string cleanSearch;
      string[] terms = StringUtils.MakeSearchTerms(search, out fullSearchString, out cleanSearch);

      List<string> unmatchedTerms = GetTermsNotInAlbumTitle(terms);

      return TracksMatching(unmatchedTerms).ToList();
    }

    private IEnumerable<Track> TracksMatching(List<string> terms)
    {
      if (terms.Count == 0)
        return Tracks;

      return Tracks.Where(track => track.Matches(terms));
    }

    private Track FirstTrackMatching(List<string> terms)
    {
      return TracksMatching(terms).FirstOrDefault();
    }

    private bool AnyTracksMatch(List<string> terms)
    {
      return (FirstTrackMatching(terms) != null);
    }

    private int GetWordScore(List<string> words, IEnumerable<string> terms, int baseScore)
    {
      int score = 0;
      foreach (string term in terms)
      {
        if (term.Contains(" "))
          continue; // no word will match

        if (words.Any(w => w.StartsWith(term)))
          score += baseScore;

        if (words.Any(w => w.Equals(term)))
          score += baseScore;
      }
      return score;
    }

    #endregion

    #region Equals and friends

    public override string ToString()
    {
      return Title;
    }

    public int CompareTo(object obj)
    {
      Album other = obj as Album;
      if (other == null)
        return 0;

      int titleCompare = this.Title.CompareTo(other.Title);
      if (titleCompare == 0)
        return this.FolderPath.CompareTo(other.FolderPath);
      return titleCompare;
    }

    public override bool Equals(object obj)
    {
      Album other = obj as Album;
      if (other == null)
        return false;
      return this.FolderPath.Equals(other.FolderPath);
    }

    public override int GetHashCode()
    {
      return FolderPath.GetHashCode();
    }

    #endregion

  }
}