using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using Airplayne.Model.Music;
using Airplayne.Model.Repository;
using Airplayne.Model.Serialization;
using Airplayne.Tools;
using System.Linq;
using Airplayne.Playback;

namespace Airplayne.Model
{
  public class PlaylistManager
  {
    #region Privates
    private LibraryManager _libraryManager;
    private PlaylistSerializer _playlistSerializer;
    private Player _player;
    protected List<Track> tracks = new List<Track>();
    private int? currentIndex;
    private int? autopilotIndex; // index of first track that's autopilot'ed
    private bool _stopAfterCurrent;
    private Playlist.AutopilotMode autopilotSetting = Playlist.AutopilotMode.Disabled;
    private bool _isLoadingState = false;
    private object _playlistLockObject = new object();
    #endregion

    public enum ListTypes { Playlist, Bag }
    public virtual ListTypes ListType { get { return ListTypes.Playlist; } }

    public TrackOpener TrackOpener { get; private set; }

    public PlaylistManager(SettingsManager settingsManager, LibraryManager libraryManager)
    {
      _libraryManager = libraryManager;
      _libraryManager.LibraryUpdated += new LibraryManager.LibraryUpdatedHandler(_libraryManager_LibraryUpdated);
      _playlistSerializer = new PlaylistSerializer(_libraryManager);
      TrackOpener = new TrackOpener(settingsManager, _libraryManager, this);
      Program.Shutdown += new Program.ProgramEvent(Program_Shutdown);
    }

    protected virtual string StateFilePath
    {
      get
      {
        return Path.Combine(DeployUtils.GetDataDirectory(), "airplayne_state.m3u");
      }
    }

    public PlaylistManager SisterListManager { get; set; }

    private void _libraryManager_LibraryUpdated()
    {
      for (int i = 0; i < Tracks.Count; i++)
        Tracks[i] = _libraryManager.GetTrack(Tracks[i].FullPath);
      DoListChanged(true, false);
    }

    #region Wrapping the player

    public class PlayerErrorEventArgs
    {
      public PlayerErrorEventArgs(bool canProceed)
      {
        CanProceed = canProceed;
      }

      public bool ProceedToNextTrack { get; set; }
      public bool CanProceed { get; private set; }
    }

    public delegate void PlayerErrorHandler(PlayerErrorEventArgs args);
    public event PlayerErrorHandler PlayerError;
    public event Player.PositionChangedHandler PlayerPositionChanged;
    public event Player.StateChangedHandler PlayerStateChanged;

    public void SetPlayer(Player player)
    {
      if (_player != null)
      {
        _player.Stop();
        _player.WaitForCompletion(TimeSpan.FromSeconds(3));

        _player.Error -= OnPlayerError;
        _player.PositionChanged -= OnPlayerPositionChanged;
        _player.StateChanged -= OnPlayerStateChanged;
        _player.TrackEnded -= Player_TrackEnded;
      }

      _player = player;

      if (_player != null)
      {
        _player.Error += OnPlayerError;
        _player.PositionChanged += OnPlayerPositionChanged;
        _player.StateChanged += OnPlayerStateChanged;
        _player.TrackEnded += Player_TrackEnded;
      }
    }

    private void OnPlayerError()
    {
      var handler = PlayerError;
      if (handler != null)
      {
        var eventArgs = new PlayerErrorEventArgs(CanNext());
        handler(eventArgs);

        if (eventArgs.CanProceed && eventArgs.ProceedToNextTrack)
        {
          if (Next())
            StartPlay();
        }
      }
    }

    private void OnPlayerPositionChanged(TimeSpan position, TimeSpan total)
    {
      var handler = PlayerPositionChanged;
      if (handler != null)
        handler(position, total);
    }

    private void OnPlayerStateChanged(Player.States state)
    {
      var handler = PlayerStateChanged;
      if (handler != null)
        handler(state);
    }

    public void SetPlayerVolume(int volume)
    {
      if (_player != null)
        _player.SetVolume(volume);
    }

    public Player.States PlayerState
    {
      get
      {
        if (_player == null)
          return Player.States.Stopped;
        else
          return _player.State;
      }
    }

    public void SetPlayerPosition(TimeSpan position)
    {
      if (_player != null)
        _player.SetPosition(position);
    }

    #endregion

    #region Delegates, events & wrapping
    public delegate void AutopilotSettingChangedHandler();
    public event AutopilotSettingChangedHandler AutopilotSettingChanged;

    public delegate void ListChangedHandler(bool listChanged, bool currentChanged);
    public delegate void PlaylistEventHandler();

    public event ListChangedHandler ListChanged;
    public event PlaylistEventHandler StopAfterCurrentChanged;

    private T FireWrapper<T>(Func<T> ar)
    {
      lock (_playlistLockObject)
      {
        List<Track> tracksWere = new List<Track>(tracks);
        int? currentIndexWas = CurrentIndex;
        Track trackWas = CurrentTrack;
        int autopilotWas = AutopilotIndex;

        T result = ar();

        if (IsCurrentTrackRemoved)
          Stop();

        if (CurrentIndex == null && AutopilotIndex > 0)
          CurrentIndex = 0;

        bool currentChanged = !object.Equals(trackWas, CurrentTrack);
        Autopilot(purge: currentChanged);

        if (autopilotWas == 0 && AutopilotIndex > 0)
          _StartPlay();

        bool listChanged = PlaylistDiffersFrom(tracksWere, currentIndexWas, autopilotWas);

        if ((listChanged || currentChanged))
          DoListChanged(listChanged, currentChanged);

        return result;
      }
    }

    private bool IsCurrentTrackRemoved
    {
      get
      {
        return CurrentIndex == null && PlayerState != Player.States.Stopped;
      }
    }

    private bool PlaylistDiffersFrom(List<Track> tracksWere, int? currentIndexWas, int? autopilotWas)
    {
      if (tracks.Count != tracksWere.Count || currentIndexWas != CurrentIndex || autopilotWas != AutopilotIndex)
        return true;

      for (int i = 0; i < tracks.Count; i++)
      {
        if (!tracks[i].Equals(tracksWere[i]))
        {
          return true;
        }
      }
      return false;
    }

    private void DoListChanged(bool listChanged, bool currentChanged)
    {
      SaveState();

      if (ListChanged != null)
        ListChanged(listChanged, currentChanged);
    }

    protected void FireWrapper(Action a)
    {
      FireWrapper<object>(delegate() { a(); return new object(); });
    }

    #endregion

    #region State properties
    public bool StopAfterCurrent
    {
      set
      {
        _stopAfterCurrent = value;
        if (StopAfterCurrentChanged != null)
          StopAfterCurrentChanged();
      }
      get { return _stopAfterCurrent; }
    }

    public List<Track> Tracks
    {
      get { return tracks; }
    }

    public void SetCurrentIndex(int? index)
    {
      FireWrapper(delegate() { CurrentIndex = index; });
    }

    public virtual int? CurrentIndex
    {
      get { return currentIndex; }
      protected set
      {
        if (tracks.Count == 0)
          currentIndex = null;
        else if (value < 0)
          currentIndex = 0;
        else if (value >= Tracks.Count)
          currentIndex = Tracks.Count - 1;
        else
          currentIndex = value;
      }
    }

    public int AutopilotIndex
    {
      get
      {
        if (CurrentIndex >= autopilotIndex)
          autopilotIndex = CurrentIndex + 1;

        if (!autopilotIndex.HasValue)
          autopilotIndex = tracks.Count;

        return autopilotIndex.Value;
      }
      set
      {
        FireWrapper(delegate()
        {
          autopilotIndex = value;
        });
      }
    }

    public Playlist.AutopilotMode AutopilotSetting
    {
      get { return autopilotSetting; }
      set
      {
        FireWrapper(delegate()
        {
          ClearAutopilotTracks();
          autopilotSetting = value;
        });
        DoAutopilotSettingChanged();
      }
    }

    private void DoAutopilotSettingChanged()
    {
      if (AutopilotSettingChanged != null)
        AutopilotSettingChanged();
    }

    public Track CurrentTrack
    {
      get
      {
        return CurrentIndex.HasValue ? Tracks[CurrentIndex.Value] : null;
      }
    }

    public Album CurrentAlbum
    {
      get
      {
        var track = CurrentTrack;
        return (track == null ? null : track.AlbumInstance);
      }
    }

    #endregion

    #region Autopilot

    private void Autopilot(bool purge = true)
    {
      AddPilotTracks();
      if (purge)
        AutoPurge();
    }

    private void AddPilotTracks()
    {
      int index = CurrentIndex ?? -1;

      while (tracks.Count - index < 4)
      {
        IMusic addition = GetAutopilotTracks();

        if (addition != null)
          tracks.AddRange(addition.Tracks);
        else
          return;
      }
    }

    private IMusic GetAutopilotTracks()
    {
      switch (autopilotSetting)
      {
        case Playlist.AutopilotMode.AddAlbums:
          return _libraryManager.RandomAlbum;
        case Playlist.AutopilotMode.AddTracks:
          return _libraryManager.RandomTrack;
        default:
          return null;
      }
    }

    private void AutoPurge()
    {
      if (CurrentIndex > 3)
      {
        List<int> toRemove = new List<int>();
        for (int i = 0; i < (CurrentIndex.Value - 3); i++)
          toRemove.Add(i);
        _RemoveTrackRange(toRemove);
      }
    }

    #endregion

    #region Playback controls

    public void Prev()
    {
      FireWrapper(delegate() { _Prev(); });
    }

    private void _Prev()
    {
      if (!CurrentIndex.HasValue)
        CurrentIndex = 0;
      else
      {
        if (tracks.Count > 0)
        {
          if (CurrentIndex - 1 >= 0)
          {
            StopAfterCurrent = false;

            CurrentIndex--;

            if (PlayerState == Player.States.Playing || PlayerState == Player.States.Loading)
              StartPlay();
            else if (PlayerState == Player.States.Paused)
              Stop();
          }
        }
      }
    }

    public void StartPlay()
    {
      FireWrapper(delegate() { _StartPlay(); });
    }

    private void _StartPlay()
    {
      if (Tracks.Count > 0)
      {
        StopAfterCurrent = false;

        if (!CurrentIndex.HasValue)
          CurrentIndex = 0;

        if (_player != null)
          _player.StartPlaying(Tracks[CurrentIndex.Value].FullPath);
      }
    }

    public void PlayPause()
    {
      if (PlayerState == Player.States.Stopped)
      {
        StartPlay();
      }
      else if (_player.State == Player.States.Playing || _player.State == Player.States.Paused)
      {
        if (_player != null)
          _player.PlayPause();
      }
    }

    public void Play()
    {
      if (PlayerState == Player.States.Paused)
      {
        PlayPause();
      }
      else
      {
        StartPlay();
      }
    }

    public void Stop()
    {
      StopAfterCurrent = false;

      if (_player != null)
        _player.Stop();
    }

    public bool Next()
    {
      return FireWrapper<bool>(() => _Next());
    }

    private bool _Next()
    {
      if (!CanNext())
        return false;

      StopAfterCurrent = false;

      if (!CurrentIndex.HasValue)
        CurrentIndex = 0;
      else
        CurrentIndex++;

      if (PlayerState != Player.States.Stopped)
        _StartPlay();

      return true;
    }

    private bool CanNext()
    {
      return tracks.Count > (CurrentIndex.HasValue ? CurrentIndex.Value + 1 : 0);
    }

    private string Player_TrackEnded()
    {
      return FireWrapper<string>(delegate
      {
        if (CurrentIndex.HasValue && tracks.Count > CurrentIndex + 1)
        {
          CurrentIndex++;
          if (!_stopAfterCurrent)
            return CurrentTrack.FullPath;

          StopAfterCurrent = false;
        }
        else
        {
          _Clear();
        }
        return null;
      });
    }
    #endregion

    #region List manipulation

    public void InsertTrack(
      Track track,
      int? index = null,
      bool asNext = false)
    {
      InsertTrackRange(new List<Track>(1) { track }, index, asNext);
    }

    public void InsertTrackRange(
      IEnumerable<Track> tt,
      int? index = null,
      bool asNext = false)
    {
      FireWrapper(delegate() { _InsertTrackRange(tt, index, asNext: asNext, beforeAutoPilot: true); });
    }

    private void _InsertTrackRange(
      IEnumerable<Track> tt,
      int? index = null,
      bool asNext = false,
      bool beforeAutoPilot = true)
    {
      bool asCurrent;
      int targetIndex = GetTargetIndex(index, asNext, beforeAutoPilot, out asCurrent);

      List<Track> reversed = new List<Track>(tt);
      reversed.Reverse();

      foreach (Track t in reversed)
        _InsertTrack(targetIndex, t);

      if (asCurrent)
        CurrentIndex = targetIndex;
    }

    private void _InsertTrack(int index, Track t)
    {
      if (index < 0 || index > tracks.Count)
        throw new ArgumentOutOfRangeException();

      tracks.Insert(index, t);
      if (index <= CurrentIndex)
        CurrentIndex++;

      if (index <= autopilotIndex)
        autopilotIndex++;
    }

    private int GetTargetIndex(int? index, bool asNext, bool beforeAutoPilot, out bool asCurrent)
    {
      if (!index.HasValue)
        index = AutopilotIndex;

      asCurrent = false;
      if (asNext)
      {
        if (!CurrentIndex.HasValue)
          index = 0;
        else if (PlayerState == Player.States.Stopped)
        {
          index = CurrentIndex.Value;
          asCurrent = true;
        }
        else
          index = CurrentIndex.Value + 1;
      }

      if (beforeAutoPilot)
        index = Math.Min(index.Value, AutopilotIndex);

      return index.Value;
    }

    public void RemoveTrack(int index)
    {
      FireWrapper(delegate() { _RemoveTrack(index); });
    }

    private void _RemoveTrack(int index)
    {
      if (index < 0 || index >= tracks.Count)
        return;

      tracks.RemoveAt(index);

      if (index == CurrentIndex)
        CurrentIndex = null;
      else if (index < CurrentIndex)
        CurrentIndex--;

      if (index < autopilotIndex)
        autopilotIndex--;
    }

    public void RemoveTrackRange(List<int> indices)
    {
      FireWrapper(delegate() { _RemoveTrackRange(indices); });
    }

    private void _RemoveTrackRange(List<int> indices)
    {
      indices.Sort();
      indices.Reverse();
      foreach (int i in indices)
      {
        _RemoveTrack(i);
      }
    }

    public void MoveTracks(List<int> indices, int? dropIndex = null, bool asNext = false)
    {
      FireWrapper(delegate() { _MoveTracks(indices, dropIndex, asNext); });
    }

    public virtual void ActivateTracks(List<int> indices, bool shift = false)
    {
      if (indices == null || indices.Count == 0)
        return;

      SetCurrentIndex(indices[0]);
      StartPlay();
    }

    public void MoveTracksToSisterList(List<int> indices, bool asNext = false)
    {
      var tt = indices.Select(index => tracks[index]).ToList();
      SisterListManager.InsertTrackRange(tt, asNext: asNext);
      RemoveTrackRange(indices);
    }

    private void _MoveTracks(List<int> indices, int? dropIndex, bool asNext)
    {
      List<Track> tt = new List<Track>();
      int? resetIndex = null;
      for (int i = 0; i < indices.Count; i++)
      {
        int index = indices[i];
        tt.Add(Tracks[index]);

        if (index < dropIndex && dropIndex > 0)
          dropIndex--;

        if (index == CurrentIndex)
          resetIndex = dropIndex + tt.Count - 1;
      }

      _RemoveTrackRange(indices);

      _InsertTrackRange(tt, dropIndex, asNext: asNext, beforeAutoPilot: false);

      if (autopilotIndex < dropIndex + tt.Count)
        autopilotIndex = dropIndex + tt.Count;

      if (resetIndex.HasValue)
        CurrentIndex = resetIndex;
    }

    public void Clear() // I hear this one's popular at parties
    {
      FireWrapper(delegate()
      {
        _Clear();
      });
    }

    protected void _Clear()
    {
      tracks.Clear();
      CurrentIndex = null;
      autopilotIndex = null;
    }

    #endregion

    #region Sorting
    private delegate List<Track> ListSortDelegate(List<Track> input);

    private void SortPlaylist(ListSortDelegate sorter)
    {
      FireWrapper(delegate()
      {
        int from = 0;
        int to = tracks.Count;

        if (CurrentIndex.HasValue)
        {
          if (PlayerState == Player.States.Stopped)
            from = CurrentIndex.Value;
          else
            from = CurrentIndex.Value + 1;
        }

        if (autopilotIndex.HasValue)
          to = autopilotIndex.Value;

        if (to > from)
        {
          List<Track> tracksToSort = tracks.GetRange(from, to - from);
          List<Track> sortedTracks = sorter(tracksToSort);

          List<int> indicesToRemove = new List<int>(sortedTracks.Count);
          for (int i = from; i < to; i++)
            indicesToRemove.Add(i);

          _RemoveTrackRange(indicesToRemove);

          _InsertTrackRange(sortedTracks, from);
        }
      });
    }

    public void RandomizePlaylist()
    {
      SortPlaylist(delegate(List<Track> input)
      {
        List<Track> shuffledTracks = new List<Track>(input.Count);
        while (input.Count > 0)
        {
          int r = Utils.Random.Next(input.Count);
          shuffledTracks.Add(input[r]);
          input.RemoveAt(r);
        }
        return shuffledTracks;
      });
    }

    public void OrderPlaylist()
    {
      SortPlaylist(delegate(List<Track> input)
      {
        input.Sort();
        return input;
      });
    }

    public void ReversePlaylist()
    {
      SortPlaylist(delegate(List<Track> input)
      {
        input.Reverse();
        return input;
      });
    }
    #endregion

    public Playlist GetPlaylist()
    {
      return new Playlist()
      {
        Tracks = this.Tracks,
        CurrentIndex = this.CurrentIndex,
        AutopilotIndex = this.AutopilotIndex,
        AutopilotSetting = this.autopilotSetting
      };
    }

    public virtual void SetPlaylist(Playlist playlist)
    {
      FireWrapper(delegate()
      {
        Stop();
        _Clear();
        tracks = playlist.Tracks;
        CurrentIndex = playlist.CurrentIndex;
        autopilotIndex = playlist.AutopilotIndex;
        if (playlist.AutopilotSetting.HasValue)
        {
          autopilotSetting = playlist.AutopilotSetting.Value;
          DoAutopilotSettingChanged();
        }
      });
    }

    #region Setup and stuff

    private void ClearAutopilotTracks()
    {
      if (autopilotIndex.HasValue)
      {
        while (tracks.Count > autopilotIndex)
          _RemoveTrack(tracks.Count - 1);
      }
    }

    public void Program_Shutdown()
    {
      SaveState();
      Stop();
      if (_player != null)
        _player.WaitForCompletion(TimeSpan.FromSeconds(3));
    }

    private void SaveState()
    {
      if (!_isLoadingState)
      {
        try
        {
          _playlistSerializer.SavePlaylist(GetPlaylist(), StateFilePath);
        }
        catch (Exception ex)
        {
          Logger.Write(ex, "Failed saving state");
        }
      }
    }

    public void LoadState()
    {
      _isLoadingState = true;

      if (File.Exists(StateFilePath))
      {
        try
        {
          SetPlaylist(_playlistSerializer.LoadPlaylist(StateFilePath));
        }
        catch (Exception ex)
        {
          Logger.Write(ex, "Failed loading state");
        }
      }

      _isLoadingState = false;
    }

    #endregion

  }

}