using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Airplayne.Model.Music;
using Airplayne.Model.Serialization;
using Airplayne.Model.Tags;
using Airplayne.Tools;

namespace Airplayne.Model.Repository
{
  public class LibraryManager
  {
    private SettingsManager _settings;

    public LibraryManager(SettingsManager settings)
    {
      _settings = settings;
    }

    #region Privates
    private Library library = new Library();
    public delegate void LibraryUpdatedHandler();
    public event LibraryUpdatedHandler LibraryUpdated;
    private string CACHE_PATH = DeployUtils.GetDataDirectory()
      + Path.DirectorySeparatorChar + "airplayne_librarycache_v3";
    #endregion

    #region Gimme music

    public Album RandomAlbum
    {
      get
      {
        return RandomAlbums(1).FirstOrDefault();
      }
    }

    public IEnumerable<Album> RandomAlbums(int? maxNum = null)
    {
      List<Album> bucket = new List<Album>(library.Albums);
      int returned = 0;
      while (bucket.Count > 0 && (!maxNum.HasValue || returned < maxNum.Value))
      {
        Album a = bucket[Utils.Random.Next(bucket.Count)];
        yield return a;
        bucket.Remove(a);
      }
    }

    public Track RandomTrack
    {
      get
      {
        if (library.Tracks.Count > 0)
        {
          return library.Tracks[Utils.Random.Next(library.Tracks.Count)];
        }
        return null;
      }
    }

    public Album GetAlbum(int albumId)
    {
      if (library.AlbumsById.ContainsKey(albumId))
        return library.AlbumsById[albumId];
      else
        return null;
    }

    public Track GetTrack(int trackId)
    {
      if (library.TracksById.ContainsKey(trackId))
        return library.TracksById[trackId];
      else
        return null;
    }

    public Track GetTrack(string fullPath)
    {
      string key = fullPath.ToLower();
      if (library.TracksByPath.ContainsKey(key))
        return library.TracksByPath[key];
      else
        return BuildTrackFromPath(fullPath);
    }

    private Track BuildTrackFromPath(string absolutePath) // hackfest
    {
      string folder = Path.GetDirectoryName(absolutePath);
      string file = Path.GetFileName(absolutePath);
      string albumTitle = null;

      foreach (string libraryRoot in _settings.Folders)
      {
        if (folder.StartsWith(libraryRoot))
        {
          albumTitle = LibraryUtils.MakeAlbumTitle(libraryRoot, folder);
          break;
        }
      }

      if (albumTitle == null)
        albumTitle = new FileInfo(absolutePath).Directory.Name;

      var tracks = new List<Track>{
        new Track(file)
      };

      Album album = new Album(folder, albumTitle, tracks); // sets album on the track      

      return tracks[0];
    }

    #endregion

    #region Searching
    private List<Album> lastSearchResult = null;
    private string lastSearchString = null;

    public List<Album> Search(string search, int? maxResults = null)
    {
      string fullSearchString, cleanSearch;
      string[] terms = StringUtils.MakeSearchTerms(search, out fullSearchString, out cleanSearch);

      List<Album> searchBucket = GetSearchBucket(cleanSearch);
      List<Album> resultList = Search(searchBucket, terms, fullSearchString);
      SetSearchBucket(resultList, cleanSearch);

      return (maxResults.HasValue ? resultList.Take(maxResults.Value).ToList() : resultList);
    }

    private List<Album> Search(List<Album> searchBucket, string[] terms, string fullSearchString)
    {
      var matches =
        from album in searchBucket
        where album.Matches(terms)
        orderby album.MatchScore(terms, fullSearchString) descending
        select album;

      return matches.ToList();
    }

    private List<Album> GetSearchBucket(string cleanSearch)
    {
      if (!string.IsNullOrEmpty(lastSearchString)
        && cleanSearch.StartsWith(lastSearchString)
        && lastSearchResult != null)
        return lastSearchResult;
      else
        return library.Albums;
    }

    private void SetSearchBucket(List<Album> result, string cleanSearch)
    {
      if (result != null)
      {
        lastSearchResult = new List<Album>(result);
        lastSearchResult.Sort();
      }
      else
        lastSearchResult = null;
      lastSearchString = cleanSearch;
    }

    #endregion

    #region Scanning
    public void ScanInBackground()
    {
      Scanner scanner = new Scanner(_settings.Folders, (result) => SaveLibrary(result.Library),
        Scanner.ScanType.BuildLibrary, _settings, this);
      scanner.StartScan();
    }

    private void SaveLibrary(Library newLibrary)
    {
      library = newLibrary;
      SetSearchBucket(null, null);
      DoLibraryUpdated();
      SaveLibraryToDisk();
    }

    private void DoLibraryUpdated()
    {
      if (LibraryUpdated != null)
        LibraryUpdated();
    }
    #endregion

    #region Serialization
    private void SaveLibraryToDisk()
    {
      LibrarySerializer.Save(library);
    }

    public void LoadLibraryFromDisk()
    {
      Library cachedLibrary = LibrarySerializer.Load();
      if (cachedLibrary != null)
      {
        library = cachedLibrary;
        DoLibraryUpdated();
      }
      else
      {
        ScanInBackground();
      }
    }
    #endregion

  }
}