﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.IO;
using Airplayne.Model.Music;

namespace Airplayne.Model.Repository
{
  public class Scanner
  {
    #region Concurrency handling (one at a time plz)
    private static List<Scanner> runningScanners = new List<Scanner>();
    private static void AddScanner(Scanner scanner)
    {
      StopAllScanners();

      lock (runningScanners)
        runningScanners.Add(scanner);
      scanner.ScanEnded += scanner_ScanEnded;
    }

    private static void scanner_ScanEnded(Scanner scanner)
    {
      scanner.ScanEnded -= scanner_ScanEnded;
      lock (runningScanners)
      {
        runningScanners.Remove(scanner);
      }
    }

    private static void StopAllScanners()
    {
      lock (runningScanners)
        foreach (Scanner scanner in runningScanners)
          scanner.StopScan();
    }
    #endregion

    #region Events
    public delegate void ScanCallback(ScanResult result);
    public delegate void ScanProgressHandler(bool done, int albums, int tracks, int? percentEstimate);
    public static event ScanProgressHandler ScanProgress;
    public enum ScanType { BuildLibrary, AdHoc }

    private delegate void ScanEndedHandler(Scanner scanner);
    private event ScanEndedHandler ScanEnded;

    #endregion

    #region Privates
    private LibraryManager _libraryManager;
    private List<string> folders;
    private List<string> extensions;
    private ScanCallback callback;
    private ScanType scanType;
    private bool stopRequested;

    private List<Album> libraryAlbumTarget = new List<Album>();
    private List<Track> adHocTrackTarget = new List<Track>();

    private int foundTracksCount = 0;
    private int foundAlbumsCount = 0;
    #endregion

    #region Scanning
    public Scanner(List<string> folders, ScanCallback callback, ScanType scanType, SettingsManager settings,
      LibraryManager libraryManager)
    {
      _libraryManager = libraryManager;
      extensions = settings.Extensions;
      this.folders = folders;
      this.callback = callback;
      this.scanType = scanType;
      AddScanner(this);
    }

    public void StartScan()
    {
      DoScanProgress(false);

      Thread scanThread = new Thread(
        () =>
        {
          ScanFolders(folders);

          if (!stopRequested)
          {
            ScanResult result = null;
            switch (scanType)
            {
              case ScanType.BuildLibrary:
                result = new ScanResult(libraryAlbumTarget);
                break;
              case ScanType.AdHoc:
                result = new ScanResult(adHocTrackTarget);
                break;
            }

            DoScanProgress(true);
            callback(result);
          }

          DoScanEnded();
        });
      scanThread.IsBackground = true;
      scanThread.Start();
    }

    private void DoScanEnded()
    {
      if (ScanEnded != null)
        ScanEnded(this);
    }

    public void StopScan()
    {
      stopRequested = true;
    }

    private void ScanFolders(List<string> folders)
    {
      foreach (string folder in folders)
      {
        if (stopRequested)
          return;

        DirectoryInfo di = new DirectoryInfo(folder);
        ScanFolder(di, di);
      }
      libraryAlbumTarget.Sort();
    }

    private void ScanFolder(DirectoryInfo topFolder, DirectoryInfo folder)
    {
      if (stopRequested)
        return;

      try
      {
        var folders = folder.EnumerateDirectories();
        foreach (DirectoryInfo di in folders)
        {
          ScanFolder(topFolder, di);
        }
        ScanForTracks(topFolder, folder);
      }
      catch (UnauthorizedAccessException)
      {
        return;
      }
      catch (DirectoryNotFoundException)
      {
        return;
      }
    }

    private void ScanForTracks(DirectoryInfo topFolder, DirectoryInfo folder)
    {
      if (stopRequested)
        return;

      List<Track> tracks = GetTracks(folder);

      if (tracks.Count > 0)
      {
        AddFoundTracks(topFolder, folder, tracks);
        DoScanProgress(false);
      }
    }

    private void AddFoundTracks(DirectoryInfo topFolder, DirectoryInfo folder, List<Track> tracks)
    {
      switch (scanType)
      {
        case ScanType.BuildLibrary:
          Album a = new Album(folder.FullName, LibraryUtils.MakeAlbumTitle(topFolder.FullName, folder.FullName), tracks);
          libraryAlbumTarget.Add(a);
          break;
        case ScanType.AdHoc:
          adHocTrackTarget.AddRange(tracks);
          break;
      }

      foundAlbumsCount++;
      foundTracksCount += tracks.Count;
    }

    private List<Track> GetTracks(DirectoryInfo folder)
    {
      var tracks =
        folder.EnumerateFiles()
        .Where(file => 
            !string.IsNullOrEmpty(file.Extension)
            && extensions.Contains(file.Extension.ToLower().Substring(1)))
        .Select(file => GetTrack(file))
        .ToList();

      tracks.Sort();
      return tracks;
    }

    private Track GetTrack(FileInfo file)
    {
      switch (scanType)
      {
        case ScanType.BuildLibrary:
          return new Track(file.Name);
        case ScanType.AdHoc:
          return _libraryManager.GetTrack(file.FullName);
        default:
          throw new ArgumentException("ScanType");
      }
    }

    private void DoScanProgress(bool done)
    {
      if (ScanProgress != null && (done || (foundAlbumsCount % 10 == 0)))
      {
        ScanProgress(done, foundAlbumsCount, foundTracksCount, null);
      }
    }

    #endregion


    public class ScanResult
    {
      private List<Track> tracks;

      public ScanResult(List<Album> libraryAlbums)
      {
        Library = new Library(libraryAlbums);
      }

      public ScanResult(List<Track> adHocTracks)
      {
        tracks = adHocTracks;
      }

      public Library Library { get; set; }
      public bool HasLibrary { get { return Library != null; } }

      public List<Track> Tracks
      {
        get
        {
          return HasLibrary ? Library.Tracks : tracks;
        }
      }
    }
  }
}