﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Airplayne.Tools;
using Airplayne.Playback;
using Airplayne.Model.Music;
using Airplayne.Model.Serialization;

namespace Airplayne.Model.Repository
{
  public class TrackOpener
  {
    private SettingsManager _settingsManager;
    private LibraryManager _libraryManager;
    private PlaylistManager _playlistManager;
    private PlaylistSerializer _playlistSerializer;
    private DateTime _lastPlaylistUpdate = DateTime.MinValue;

    public TrackOpener(SettingsManager settingsManager, LibraryManager libraryManager, PlaylistManager playlistManager)
    {
      _settingsManager = settingsManager;
      _libraryManager = libraryManager;
      _playlistManager = playlistManager;
      _playlistManager.ListChanged += new PlaylistManager.ListChangedHandler(_playlistManager_ListChanged);
      _playlistSerializer = new PlaylistSerializer(_libraryManager);
    }

    void _playlistManager_ListChanged(bool listChanged, bool currentChanged)
    {
      if (listChanged)
        _lastPlaylistUpdate = DateTime.Now;
    }

    public void AddTracks(params string[] paths)
    {
      AddTracks(targetIndex: null, paths: paths);
    }

    public void AddTracks(int? targetIndex, params string[] paths)
    {
      List<string> folders;
      List<string> playlists;
      List<string> files;
      OrganizePaths(paths, out folders, out playlists, out files);

      if (folders.Count == 0
        && files.Count == 0
        && playlists.Count == 1
        && _playlistManager.Tracks.Count == 0)
        OpenPlaylist(playlists[0]);
      else
        AddTracks(folders, playlists, files, targetIndex);
    }

    private void AddTracks(List<string> folders, List<string> playlists, List<string> files, int? targetIndex)
    {
      var tracks = new List<Track>();

      foreach (string file in files)
        tracks.Add(GetTrack(file));
      foreach (string playlist in playlists)
        tracks.AddRange(GetTracks(playlist));

      if (folders.Count > 0)
      {
        Scanner folderScanner = new Scanner(
          folders,
          scanResult => AfterScan(scanResult, tracks, targetIndex, _lastPlaylistUpdate),
          Scanner.ScanType.AdHoc,
          _settingsManager,
          _libraryManager);
        folderScanner.StartScan();
      }
      else
        _playlistManager.InsertTrackRange(tracks, targetIndex);
    }

    private void AfterScan(Scanner.ScanResult scanResult, List<Track> otherTracks, int? targetIndex, DateTime playlistUpdate)
    {
      otherTracks.InsertRange(0, scanResult.Tracks);

        _playlistManager.InsertTrackRange(
          otherTracks,
          (_lastPlaylistUpdate == playlistUpdate) ? targetIndex : null
        );
    }

    private List<Track> GetTracks(string playlistPath)
    {
      Playlist playlist = _playlistSerializer.LoadPlaylist(playlistPath);
      return playlist.NonAutopilotTracks;
    }

    private Track GetTrack(string file)
    {
      return _libraryManager.GetTrack(file);
    }

    public void OpenPlaylist(string path)
    {
      Playlist playlist = _playlistSerializer.LoadPlaylist(path);
      _playlistManager.SetPlaylist(playlist);
    }

    public void SavePlaylist(Playlist playlist, string path)
    {
      _playlistSerializer.SavePlaylist(playlist, path);
    }

    private void OrganizePaths(string[] paths, out List<string> folders, out List<string> playlists, out List<string> files)
    {
      Array.Sort(paths);
      folders = new List<string>();
      playlists = new List<string>();
      files = new List<string>();

      foreach (string path in paths)
      {
        if (Directory.Exists(path))
          folders.Add(path);
        else if (File.Exists(path))
        {
          string extension = GetExtension(path);
          if (extension == "m3u")
            playlists.Add(path);
          else if (_settingsManager.Extensions.Contains(extension))
            files.Add(path);
        }
      }
    }

    private string GetExtension(string path)
    {
      return (Path.GetExtension(path) ?? string.Empty).ToLower().TrimStart('.');
    }

  }
}
