using System;
using System.Collections.Generic;
using System.Linq;
using Airplayne.Tools.Threading;
using Airplayne.Playback;
using Airplayne.Interface.Hotkeys;
using System.Windows.Forms;

namespace Airplayne.Model
{
  public class SettingsManager
  {
    private Properties.Settings _default = Properties.Settings.Default;
    private RequestQueue<object, object> _saveQueue;
    private List<PlayerDescription> _availablePlayers;

    #region Events

    public delegate void SettingChangedHandler();
    public event SettingChangedHandler LibrarySettingsChanged;
    public event SettingChangedHandler HotkeySettingsChanged;
    public delegate void VolumeChangedHandler(int volume);
    public event VolumeChangedHandler VolumeChanged;
    public delegate void PlayerTypeChangedHandler(PlayerDescription playerDescription);
    private PlayerTypeChangedHandler _playerTypeChangedHandler;

    #endregion

    public SettingsManager(PlayerTypeChangedHandler playerTypeChangedHandler)
    {
      _playerTypeChangedHandler = playerTypeChangedHandler;
      _saveQueue = new RequestQueue<object, object>((parameter) =>
      {
        SaveSettings();
        return null;
      }, threadPriority: System.Threading.ThreadPriority.BelowNormal);
      InitAvailablePlayers();
    }

    private void SaveSettings()
    {
      _default.Save();
    }

    private void RequestSave()
    {
      _saveQueue.Request(null);
    }

    public List<string> Folders
    {
      get
      {
        string setting = _default.Folders;
        if (string.IsNullOrEmpty(setting))
          return new List<string>();

        return setting.Split(';').ToList();
      }
      set
      {
        List<string> foldersWere = Folders;
        string[] arr = value.ToArray();
        string setting = string.Join(";", arr);
        _default.Folders = setting;
        RequestSave();

        bool changed = false;
        if (foldersWere.Count != value.Count)
          changed = true;
        else
          for (int i = 0; i < value.Count; i++)
          {
            if (foldersWere[i] != value[i])
            {
              changed = true;
              break;
            }
          }

        if (changed && LibrarySettingsChanged != null)
          LibrarySettingsChanged();
      }
    }

    public List<string> Extensions
    {
      get
      {
        string setting = _default.Extensions;
        if (string.IsNullOrEmpty(setting))
          return new List<string>()
          {
            "mp3",
            "wma",
            "ogg",
            "flac",
            "aac",
            "m4a",
            "mpc"
          };

        return setting.Split(';').ToList();
      }
      set
      {
        string[] arr = value.ToArray();
        string setting = string.Join(";", arr).ToLower();
        if (_default.Extensions != setting)
        {
          _default.Extensions = setting;
          RequestSave();
          if (LibrarySettingsChanged != null)
            LibrarySettingsChanged();
        }
      }
    }

    public int Volume
    {
      get { return _default.Volume; }
      set
      {
        value = Math.Max(0, value);
        value = Math.Min(100, value);
        _default.Volume = value;
        RequestSave();
        if (VolumeChanged != null)
          VolumeChanged(value);
      }
    }

    public List<HotkeyAssignment> Hotkeys
    {
      get
      {
        string setting = _default.Hotkeys;
        var hotkeys = string.IsNullOrEmpty(setting) ? DefaultHotkeys : DeserializeHotkeys(setting);
        return hotkeys
            .OrderBy(hotkey => hotkey.HotkeyAction)
            .ToList();
      }
      set
      {
        _default.Hotkeys = SerializeHotkeys(value);
        RequestSave();
        if (HotkeySettingsChanged != null)
          HotkeySettingsChanged();
      }
    }

    private string SerializeHotkeys(List<HotkeyAssignment> hotkeys)
    {
      return string.Join(";", hotkeys.Select(assignment => assignment.Serialize()).ToArray());
    }

    private List<HotkeyAssignment> DeserializeHotkeys(string setting)
    {
      return setting.Split(';')
        .Select(assignmentString => HotkeyAssignment.DeSerialize(assignmentString))
        .ToList();
    }

    private List<HotkeyAssignment> DefaultHotkeys
    {
      get
      {
        return new List<HotkeyAssignment>()
        {
          new HotkeyAssignment(new Hotkey(Hotkey.Modifiers.None, Keys.MediaPreviousTrack), HotkeyAssignment.HotkeyActions.Previous),
          new HotkeyAssignment(new Hotkey(Hotkey.Modifiers.None, Keys.MediaPlayPause), HotkeyAssignment.HotkeyActions.PlayPause),
          new HotkeyAssignment(new Hotkey(Hotkey.Modifiers.None, Keys.MediaStop), HotkeyAssignment.HotkeyActions.Stop),
          new HotkeyAssignment(new Hotkey(Hotkey.Modifiers.None, Keys.MediaNextTrack), HotkeyAssignment.HotkeyActions.Next)
        };
      }
    }

    #region Player settings

    public PlayerDescription PlayerType
    {
      get
      {
        return _availablePlayers.Where(player => player.Name == _default.PlayerType).First();
      }
      set
      {
        _playerTypeChangedHandler(value);
        _default.PlayerType = value.Name;
        RequestSave();
      }
    }

    public IEnumerable<PlayerDescription> AvailablePlayersByPriority
    {
      get
      {
        return _availablePlayers
          .OrderByDescending(player => (player.Name == _default.PlayerType))
          .ThenByDescending(player => player.Awesomeness);
      }
    }

    public IEnumerable<PlayerDescription> AvailablePlayersByName
    {
      get
      {
        return _availablePlayers.OrderBy(player => player.Name);
      }
    }

    public class PlayerDescription
    {
      public Type ImplementingType { get; set; }
      public string Name { get; set; }
      public int Awesomeness { get; set; }

      public override string ToString()
      {
        return Name;
      }
    }

    private void InitAvailablePlayers()
    {
      _availablePlayers = new List<PlayerDescription>()
        {
          new PlayerDescription()
          {
            ImplementingType = typeof(BassWrapper),
            Name = "BASS Audio Library",
            Awesomeness = 400
          },
          new PlayerDescription() 
          {
            ImplementingType = typeof(VlcWrapper),
            Name = "VLC Media Player",
            Awesomeness = 100
          },
          new PlayerDescription()
          {
            ImplementingType = typeof(WmpWrapper),
            Name = "Windows Media Player",
            Awesomeness = 4
          }
        };
    }

    #endregion
  }
}
