﻿using System;
using MediaInfoLib;

namespace Airplayne.Model.Tags
{
  public class MediaInfoTagReader : ITagReader
  {
    private MediaInfo _mediaInfo;

    public MediaInfoTagReader()
    {
      _mediaInfo = new MediaInfo();
    }

    public TagInfo GetTagInfo(string filePath)
    {
      using (var mediaInfo = new MediaInfoWrapper(filePath))
      {
        return new TagInfo()
        {
          AlbumTitle = mediaInfo.GetAlbumTitle(),
          Artist = mediaInfo.GetArtist(),
          Duration = ParseDuration(mediaInfo.GetDuration()),
          Title = mediaInfo.GetTitle(),
          TrackNumber = ParseTrackNumber(mediaInfo.GetTrackNumber()),
          Year = ParseYear(mediaInfo.GetYear())
        };
      }
    }

    private TimeSpan? ParseDuration(string durationString)
    {
      int? ms = TryParseInt(durationString);
      if (!ms.HasValue)
        return null;
      try
      {
        return TimeSpan.FromMilliseconds(ms.Value);
      }
      catch (Exception ex)
      {
        if (ex is OverflowException || ex is ArgumentException)
          return null;
        else
          throw;
      }
    }

    private int? ParseTrackNumber(string trackString)
    {
      return TryParseInt(trackString);
    }

    private int? ParseYear(string yearString)
    {
      return TryParseInt(yearString);
    }

    private static int? TryParseInt(string str)
    {
      int parsedYear;
      if (int.TryParse(str, out parsedYear))
        return parsedYear;
      else
        return null;
    }




    private class MediaInfoWrapper : IDisposable
    {
      private static MediaInfo _mediaInfo;

      public MediaInfoWrapper(string filePath)
      {
        if (_mediaInfo == null)
        {
          _mediaInfo = new MediaInfo();
          _mediaInfo.Option("Internet", "No");
        }

        if (_mediaInfo.Open(filePath) != 1)
          throw new TagReaderException();
      }

      public string GetArtist()
      {
        return Inform("General;%Performer%");
      }

      public string GetTitle()
      {
        return Inform("General;%Track%");
      }

      public string GetAlbumTitle()
      {
        return Inform("General;%Album%");
      }

      public string GetTrackNumber()
      {
        return Inform("General;%Track/Position%");
      }

      public string GetYear()
      {
        return Inform("General;%Recorded_Date%");
      }

      public string GetDuration()
      {
        return Inform("General;%Duration%");
      }

      /// <summary>
      /// See _mediaInfo.Option("Info_Parameters") for list of options
      /// </summary>
      /// <param name="template"></param>
      /// <returns></returns>
      private static string Inform(string template)
      {
        _mediaInfo.Option("Inform", template);
        return _mediaInfo.Inform();
      }

      public void Dispose()
      {
        _mediaInfo.Close();
      }
    }

  }
}
