﻿using System;
using System.IO;
using Airplayne.Tools;
using Un4seen.Bass;
using Un4seen.Bass.AddOn.Fx;

namespace Airplayne.Playback
{
  public class BassWrapper : Player
  {
    private int? _channel = null;
    private int _volumeFx;
    private int _volume;
    private SYNCPROC _channelEndedHandler;
    protected override TimeSpan VolumeLag { get { return TimeSpan.FromMilliseconds(500); } }

    protected override void _Initialize()
    {
      Register();
      LoadBassDlls();

      _channelEndedHandler = new SYNCPROC(ChannelEnded);
      Bass.BASS_Init(-1, 44100, BASSInit.BASS_DEVICE_DEFAULT, IntPtr.Zero);
    }

    private void Register()
    {
      BassNet.OmitCheckVersion = true;
      if (File.Exists("BassLicense.txt"))
      {
        var lines = File.ReadAllLines("BassLicense.txt");
        BassNet.Registration(lines[0], lines[1]);
      }
      else
        Logger.Write("BassLicense.txt not present - oh noes, we'll get a splash screen!");
    }

    private static void LoadBassDlls()
    {
      string bassDirectory = Path.Combine(DeployUtils.ExecutableDirectoryPath, "Bass");
      if (!Bass.LoadMe(bassDirectory))
        throw new FileNotFoundException("Failed loading bass.dll");
      if (!BassFx.LoadMe(bassDirectory))
        throw new FileNotFoundException("Failed loading bass_fx.dll");

      Bass.BASS_PluginLoadDirectory(bassDirectory);
    }

    protected override void _Stop()
    {
      if (_channel.HasValue)
      {
        Bass.BASS_ChannelStop(_channel.Value);
        _channel = null;
        PlayerStateChanged(States.Stopped);
      }
    }

    protected override void _Resume()
    {
      if (_channel.HasValue)
      {
        Bass.BASS_ChannelPlay(_channel.Value, false);
        PlayerStateChanged(States.Playing);
      }
    }

    protected override void _Pause()
    {
      if (_channel.HasValue)
      {
        Bass.BASS_ChannelPause(_channel.Value);
        PlayerStateChanged(States.Paused);
      }
    }

    protected override TimeSpan? Position
    {
      get
      {
        if (_channel.HasValue)
        {
          return TimeSpan.FromSeconds(
            Bass.BASS_ChannelBytes2Seconds(_channel.Value, Bass.BASS_ChannelGetPosition(_channel.Value)));
        }

        return null;
      }
    }

    protected override TimeSpan? Total
    {
      get
      {
        if (_channel.HasValue)
        {
          return TimeSpan.FromSeconds(
            Bass.BASS_ChannelBytes2Seconds(_channel.Value, Bass.BASS_ChannelGetLength(_channel.Value)));
        }
        return null;
      }
    }

    protected override void _SetPosition(TimeSpan position)
    {
      if (_channel.HasValue)
      {
        Bass.BASS_ChannelSetPosition(_channel.Value, position.TotalSeconds);
      }
    }

    protected override void _StartPlaying(string path)
    {
      PlayerStateChanged(States.Loading);
      int channel = Bass.BASS_StreamCreateFile(path, 0, 0, BASSFlag.BASS_STREAM_AUTOFREE);
      if (channel == 0)
      {
        _channel = null;
        PlayerStateChanged(States.Stopped);
        OnError();
      }
      else
      {
        _channel = channel;
        Bass.BASS_ChannelSetSync(_channel.Value, BASSSync.BASS_SYNC_END, 0, _channelEndedHandler, IntPtr.Zero);
        _volumeFx = Bass.BASS_ChannelSetFX(_channel.Value, BASSFXType.BASS_FX_BFX_VOLUME, 0);
        SetChannelVolume();
        Bass.BASS_ChannelPlay(channel, true);
        PlayerStateChanged(States.Playing);
      }
    }

    private void ChannelEnded(int handle, int channel, int data, IntPtr user)
    {
      _channel = null;
      PlayerStateChanged(States.Stopped);
      string nextTrackPath = OnTrackEnded();
      if (!string.IsNullOrEmpty(nextTrackPath))
        _StartPlaying(nextTrackPath);
    }

    protected override int _Volume
    {
      get
      {
        return _volume;
      }
      set
      {
        _volume = value;
        SetChannelVolume();
      }
    }

    private void SetChannelVolume()
    {
      if (_channel.HasValue)
      {
        var volumeParameters = new BASS_BFX_VOLUME((float)_volume/100, BASSFXChan.BASS_BFX_CHANALL);
        Bass.BASS_FXSetParameters(_volumeFx, volumeParameters);
      }
    }

  }
}
