﻿using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;
using Airplayne.Tools;
using Airplayne.Tools.Threading;

namespace Airplayne.Playback
{
  public abstract class Player
  {
    #region Privates
    private object _stateLockObject = new object();
    private int _fadeLength = 500;
    private ActionQueue<PlayerAction> _actionQueue;
    private Timer _positionUpdateTimer;
    #endregion

    #region Delegates, events & enums
    public delegate void PositionChangedHandler(TimeSpan newPosition, TimeSpan total);
    public delegate void StateChangedHandler(States newState);
    public delegate string TrackEndedHandler();
    public delegate void ErrorHandler();
    public event TrackEndedHandler TrackEnded;
    public event PositionChangedHandler PositionChanged;
    public event StateChangedHandler StateChanged;
    public event ErrorHandler Error;
    public enum States { Stopped, Loading, Playing, Paused }
    #endregion

    #region Magic numbers

    protected virtual TimeSpan VolumeLag { get { return TimeSpan.Zero; } }
    protected virtual TimeSpan PrePauseLag { get { return TimeSpan.Zero; } }
    protected virtual TimeSpan PauseLag { get { return TimeSpan.Zero; } }
    protected virtual TimeSpan ResumeLag { get { return TimeSpan.Zero; } }

    #endregion

    public Player()
    {
      _actionQueue = new ActionQueue<PlayerAction>();
      _actionQueue.ActionFailed += _actionQueue_ActionFailed;
      _positionUpdateTimer = new Timer(PositionUpdateTimerCallback, null, Timeout.Infinite, Timeout.Infinite);

      InitializeOnPlayerThread();
    }

    private void InitializeOnPlayerThread()
    {
      Exception innerException = null;
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.Initialize, () =>
      {
        try
        {
          _Initialize();
        }
        catch (Exception ex)
        {
          if (ex is COMException || ex is FileNotFoundException)
            innerException = ex;
          else
            throw;
        }
      }));
      _actionQueue.WaitForCompletion(TimeSpan.FromHours(1));
      if (innerException != null)
        throw new PlayerInitializationException("Player initialization failed", innerException);
    }

    #region Public interface
    public void StartPlaying(string path)
    {
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.StartPlaying, () =>
      {
        _FadingStop();
        _StartPlaying(path);
      }));
    }

    public void PlayPause()
    {
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.PlayPause, () =>
      {
        switch (State)
        {
          case States.Paused:
            FadingResume();
            break;
          case States.Playing:
            FadingPause();
            break;
          default:
            return;
        }
      }));
    }

    public void Stop()
    {
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.Stop, () =>
      {
        _FadingStop();
      }));
    }

    public void WaitForCompletion(TimeSpan timeout)
    {
      _actionQueue.WaitForCompletion(timeout);
    }

    public States State { get; protected set; }

    public void SetVolume(int volume)
    {
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.SetVolume, () =>
      {
        _Volume = volume;
      }));
    }

    public void SetPosition(TimeSpan position)
    {
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.SetPosition, () =>
      {
        _SetPosition(position);
      }));
    }
    #endregion

    #region Abstracts
    protected abstract void _Initialize();
    protected abstract void _Stop();
    protected abstract void _Resume();
    protected abstract void _Pause();
    protected abstract TimeSpan? Position { get; }
    protected abstract TimeSpan? Total { get; }
    protected abstract void _SetPosition(TimeSpan position);
    protected abstract void _StartPlaying(string path);
    protected abstract int _Volume { get; set; }
    #endregion

    #region Event handling

    [DebuggerStepThrough]
    private void PositionUpdateTimerCallback(object state)
    {
      _actionQueue.EnqueueAction(new PlayerAction(PlayerAction.ActionTypes.GetPosition, () =>
      {
        if (State == States.Playing)
        {
          var position = Position;
          var total = Total;
          if (position.HasValue && total.HasValue)
          {
            OnPositionChanged(position.Value, total.Value);
          }
        }
      }));
    }

    private void _actionQueue_ActionFailed(PlayerAction action, Exception ex)
    {
      Logger.Write(ex);
      throw new ApplicationException("Unexpected exception in player", ex);
    }

    #endregion

    #region Event triggering

    protected string OnTrackEnded()
    {
      var handler = TrackEnded;
      if (handler != null)
        return handler();
      else
        return null;
    }

    private void OnPositionChanged(TimeSpan newPosition, TimeSpan total)
    {
      var handler = PositionChanged;
      if (handler != null)
        handler(newPosition, total);
    }

    private void OnStateChanged(States newState)
    {
      var handler = StateChanged;
      if (handler != null)
        handler(newState);
    }

    protected void OnError()
    {
      var handler = Error;
      if (handler != null)
      {
        ThreadPool.QueueUserWorkItem(state =>
        {
          handler();
        });
      }
    }

    protected void PlayerStateChanged(Player.States newState)
    {
      if (State == newState)
        return;

      lock (_stateLockObject)
      {
        State = newState;
        Monitor.PulseAll(_stateLockObject);
      }
      SetPositionTimer(newState);
      OnStateChanged(newState);
    }

    private void SetPositionTimer(States newState)
    {
      if (newState == States.Playing || newState == States.Paused)
        _positionUpdateTimer.Change(TimeSpan.FromSeconds(0.5), TimeSpan.FromSeconds(1));
      else
        _positionUpdateTimer.Change(Timeout.Infinite, Timeout.Infinite);
    }

    #endregion

    #region Private tools

    private void WaitForState(params States[] states)
    {
      lock (_stateLockObject)
      {
        while (!states.Contains(State))
          Monitor.Wait(_stateLockObject);
      }
    }

    protected void FadeOutAndExecute(Action action)
    {
      int curentVolume = _Volume;
      if (curentVolume > 0)
      {
        int sleepInterval = _fadeLength / curentVolume;
        while (_Volume > 0)
        {
          _Volume--;
          Thread.Sleep(sleepInterval);
        }
      }
      Thread.Sleep(VolumeLag);
      action();
      _Volume = curentVolume;
    }

    protected void ExecuteAndFadeIn(Action action)
    {
      int currentVolume = _Volume;
      _Volume = 0;
      action();
      if (currentVolume > 0)
      {
        int sleepInterval = _fadeLength / currentVolume;
        while (_Volume < currentVolume)
        {
          Thread.Sleep(sleepInterval);
          _Volume++;
        }
      }
    }

    private void _FadingStop()
    {
      switch (State)
      {
        case States.Playing:
          FadeOutAndExecute(() => _Stop());
          break;
        case States.Paused:
        case States.Loading:
          _Stop();
          break;
      }

      WaitForState(States.Stopped);
    }

    private void FadingResume()
    {
      ExecuteAndFadeIn(() =>
      {
        _Resume();
        WaitForState(States.Playing, States.Stopped);
        Thread.Sleep(ResumeLag);
      });
    }

    private void FadingPause()
    {
      FadeOutAndExecute(() =>
      {
        _Pause();
        WaitForState(States.Paused, States.Stopped);
        Thread.Sleep(PauseLag);
      });
    }

    #endregion

    private class PlayerAction : ActionQueue<PlayerAction>.QueuedAction
    {
      public enum ActionTypes
      {
        StartPlaying,
        PlayPause,
        Stop,
        GetPosition,
        SetPosition,
        SetVolume,
        Initialize
      }

      public ActionTypes ActionType { get; set; }

      public PlayerAction(ActionTypes actionType, Action action)
        : base(action)
      {
        ActionType = actionType;
      }

      public override bool Replaces(PlayerAction queuedAction)
      {
        switch (ActionType)
        {
          case ActionTypes.StartPlaying:
          case ActionTypes.Stop:
            switch (queuedAction.ActionType)
            {
              case ActionTypes.GetPosition:
              case ActionTypes.PlayPause:
              case ActionTypes.SetPosition:
              case ActionTypes.StartPlaying:
              case ActionTypes.Stop:
                return true;
              default:
                return false;
            }
          case ActionTypes.SetVolume:
          case ActionTypes.SetPosition:
          case ActionTypes.GetPosition:
          case ActionTypes.PlayPause:
            return (queuedAction.ActionType == ActionType);
          case ActionTypes.Initialize:
          default:
            return false;
        }
      }


    }


  }
}