﻿using System;
using AXVLC;
using System.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Microsoft.Win32;
using System.Security;
using System.IO;
using Airplayne.Tools;

namespace Airplayne.Playback
{
  public class VlcWrapper : Player
  {
    protected override TimeSpan VolumeLag { get { return TimeSpan.FromMilliseconds(400); } }
    protected override TimeSpan ResumeLag { get { return TimeSpan.FromMilliseconds(300); } }

    private VLCPlugin2 _player;
    private int? _playlistItemId = null;

    public VlcWrapper()
      : base()
    {
    }

    private void HookUpEvents()
    {
      _player.MediaPlayerPaused += () => PlayerPaused();
      _player.MediaPlayerPlaying += () => PlayerPlaying();

      _player.MediaPlayerOpening += () => PlayerLoading();
      _player.MediaPlayerEncounteredError += () => PlayerError();
      _player.MediaPlayerStopped += () => PlayerStopped();
      _player.MediaPlayerEndReached += () => PlayerTrackEnded();
    }

    private void PlayerPaused()
    {
      PlayerStateChanged(States.Paused);
    }

    private void PlayerPlaying()
    {
      PlayerStateChanged(States.Playing);
    }

    private void PlayerStopped()
    {
      PlayerStateChanged(States.Stopped);
    }

    private void PlayerTrackEnded()
    {
      PlayerStateChanged(States.Stopped);
      string nextTrackPath = OnTrackEnded();
      if (!string.IsNullOrEmpty(nextTrackPath))
        _StartPlaying(nextTrackPath);
    }

    private void PlayerLoading()
    {
      PlayerStateChanged(States.Loading);
    }

    private void PlayerError()
    {
      OnError();
      PlayerStopped();
    }

    protected override void _Initialize()
    {
      _player = new VLCPlugin2();
      HookUpEvents();
      WakeUpVlcAsync();
    }

    /// <summary>
    /// ActiveX VLC won't put out sound after a system reboot unless vlc.exe has been started. 
    /// Hack city ahoy.
    /// </summary>
    private void WakeUpVlcAsync()
    {
      Thread pokerThread = new Thread(() =>
      {
        WakeUpVlc();
      });
      pokerThread.IsBackground = false;
      pokerThread.Start();
    }

    private void WakeUpVlc()
    {
      string vlcPath = GetVlcPath();
      if (string.IsNullOrEmpty(vlcPath))
        return;

      var psInfo = new ProcessStartInfo(vlcPath)
      {
        Arguments = "--intf=none"
      };

      using (var vlcProcess = Process.Start(psInfo))
      {
        Thread.Sleep(TimeSpan.FromSeconds(4));
        vlcProcess.TryKill();
      }
    }

    private string GetVlcPath()
    {
      string vlcPath = GetVlcPathFromRegistry();
      if (string.IsNullOrEmpty(vlcPath))
        vlcPath = GetVlcPathFromProgramFiles();

      return vlcPath;
    }

    private string GetVlcPathFromProgramFiles()
    {
      string vlcPath =
        Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ProgramFilesX86),
        "VideoLAN", "VLC", "vlc.exe");
      if (File.Exists(vlcPath))
        return vlcPath;
      else
        return null;        
    }

    private static string GetVlcPathFromRegistry()
    {
      try
      {
        var vlcKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\VideoLan\VLC", false);
        if (vlcKey == null)
          return null;

        return vlcKey.GetValue(null).ToString();
      }
      catch (Exception ex)
      {
        Logger.Write(ex, "Exception getting VLC path");
        if (ex is ObjectDisposedException || ex is SecurityException || ex is IOException || ex is UnauthorizedAccessException)
          return null;
        else
          throw;
      }
    }

    protected override void _StartPlaying(string path)
    {
      if (_playlistItemId.HasValue)
        _player.playlist.items.remove(_playlistItemId.Value);
      string uri = new Uri(path).AbsoluteUri;        
      _playlistItemId = _player.playlist.add(uri);
      _player.playlist.playItem(_playlistItemId.Value);
    }

    protected override void _Resume()
    {
      _player.playlist.togglePause();
    }

    protected override void _Pause()
    {
      _player.playlist.togglePause();
    }

    protected override void _Stop()
    {
      _player.playlist.stop();
    }

    protected override int _Volume
    {
      get
      {
        return _player.Volume;
      }
      set
      {
        _player.Volume = value;
      }
    }

    protected override TimeSpan? Position
    {
      get
      {
        return TimeSpan.FromMilliseconds(_player.input.Time);
      }
    }

    protected override void _SetPosition(TimeSpan position)
    {
      _player.input.Time = position.TotalMilliseconds;
    }

    protected override TimeSpan? Total
    {
      get
      {
        return TimeSpan.FromMilliseconds(_player.input.Length);
      }
    }

  }
}