﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WMPLib;
using System.Threading;
using System.Runtime.InteropServices;
using Airplayne.Tools;

namespace Airplayne.Playback
{
  /// <summary>
  /// "Creating the Windows Media Player Control Programmatically"
  /// http://msdn2.microsoft.com/en-us/library/bb262248(VS.85).aspx
  /// </summary>
  public class WmpWrapper : Player
  {
    protected override TimeSpan PauseLag { get { return TimeSpan.FromMilliseconds(250); } }

    private WindowsMediaPlayer _wmp;
    private IWMPPlayer4 Wmp { get { return _wmp; } }
    private IWMPPlaylist _twoSongs;

    public WmpWrapper()
      : base()
    {

    }

    private void _wmp_MediaError(object pMediaObject)
    {
      OnError();
    }

    private void _wmp_PlayStateChange(int newStateInt)
    {
      var wmpState = (WMPPlayState)Enum.ToObject(typeof(WMPPlayState), newStateInt);

      switch (wmpState)
      {
        case WMPPlayState.wmppsBuffering:
        case WMPPlayState.wmppsReconnecting:
        case WMPPlayState.wmppsTransitioning:
        case WMPPlayState.wmppsScanForward:
        case WMPPlayState.wmppsScanReverse:
          PlayerStateChanged(States.Loading);
          break;
        case WMPPlayState.wmppsPaused:
          PlayerStateChanged(States.Paused);
          break;
        case WMPPlayState.wmppsPlaying:
          PlayerStateChanged(States.Playing);
          break;
        case WMPPlayState.wmppsReady:
        case WMPPlayState.wmppsStopped:
        case WMPPlayState.wmppsLast:
        case WMPPlayState.wmppsUndefined:
        case WMPPlayState.wmppsWaiting:
          PlayerStateChanged(States.Stopped);
          break;
        case WMPPlayState.wmppsMediaEnded:
          PlayerTrackEnded();
          break;
      }
    }

    private void PlayerTrackEnded()
    {
      PlayerStateChanged(States.Stopped);
      string nextTrackPath = OnTrackEnded();
      if (!string.IsNullOrEmpty(nextTrackPath))
        EnqueueNextTrack(nextTrackPath);
    }

    private void EnqueueNextTrack(string nextTrackPath)
    {
      var media = Wmp.newMedia(nextTrackPath);
      _twoSongs.appendItem(media);
      if (_twoSongs.count > 2)
        _twoSongs.removeItem(_twoSongs.get_Item(0));
    }

    protected override void _StartPlaying(string path)
    {
      IWMPMedia media = Wmp.newMedia(path);
      _twoSongs.clear();
      _twoSongs.appendItem(media);
      Wmp.controls.play();
    }

    protected override void _Initialize()
    {
      _wmp = new WindowsMediaPlayer();
      _wmp.PlayStateChange += _wmp_PlayStateChange;
      _wmp.MediaError += _wmp_MediaError;
      _twoSongs = Wmp.newPlaylist("airplayne_playlist", null);
      Wmp.currentPlaylist = _twoSongs;
    }

    protected override void _Stop()
    {
      Wmp.controls.stop();
    }

    protected override void _Pause()
    {
      Wmp.controls.pause();
    }

    protected override void _Resume()
    {
      Wmp.controls.play();
    }

    protected override TimeSpan? Position
    {
      get
      {
        return TimeSpan.FromSeconds(Wmp.controls.currentPosition);
      }
    }

    protected override void _SetPosition(TimeSpan position)
    {
      Wmp.controls.currentPosition = position.TotalSeconds;
    }

    protected override TimeSpan? Total
    {
      get
      {
        var currentMedia = Wmp.currentMedia;
        if (currentMedia != null)
          return TimeSpan.FromSeconds(currentMedia.duration);
        else
          return null;
      }
    }

    protected override int _Volume
    {
      get
      {
        return Wmp.settings.volume;
      }
      set
      {
        try
        {
          Wmp.settings.volume = value;
        }
        catch (COMException ex)
        {
          Logger.Write(ex, "Exception setting WMP volume, we'll ignore it.");
        }
      }
    }

  }
}