﻿using System;
using System.Collections.Generic;
using System.Threading;

namespace Airplayne.Tools.Threading
{
  public class ActionQueue : ActionQueue<ActionQueue.DefaultAction>
  {
    public void EnqueueAction(Action action)
    {
      base.EnqueueAction(new DefaultAction(action));
    }

    public class DefaultAction : ActionQueue<DefaultAction>.QueuedAction
    {
      public DefaultAction(Action action) : base(action) { }

      public override bool Replaces(DefaultAction queuedAction)
      {
        return false;
      }
    }
  }

  public class ActionQueue<TQeueuedAction>
    : IDisposable
    where TQeueuedAction : ActionQueue<TQeueuedAction>.QueuedAction
  {
    private bool _stopped = false;
    private object _queueLockObject = new object();
    private List<TQeueuedAction> _queue;
    private TQeueuedAction _currentAction = null;
    private Thread _thread;

    public delegate void ActionFailedHandler(TQeueuedAction action, Exception ex);
    public event ActionFailedHandler ActionFailed;

    public ActionQueue()
    {
      _queue = new List<TQeueuedAction>();
      _thread = new Thread(ExecuteQueue);
      _thread.IsBackground = true;
      _thread.Start();
    }

    public void WaitForCompletion(TimeSpan timeout)
    {
      lock (_queueLockObject)
      {
        DateTime deadline = DateTime.Now.Add(timeout);

        while ((_currentAction != null || _queue.Count > 0) && timeout > TimeSpan.Zero)
        {
          Monitor.Wait(_queueLockObject, timeout);
          timeout = deadline - DateTime.Now;
        }
      }
    }

    private void ExecuteQueue()
    {
      while (!_stopped)
      {
        WaitForAction();
        try
        {
          _currentAction.Action();
        }
        catch (Exception ex)
        {
          OnActionFailed(ex);
        }
        finally
        {
          lock (_queueLockObject)
          {
            _currentAction = null;
            Monitor.PulseAll(_queueLockObject);
          }
        }
      }
    }

    private void OnActionFailed(Exception ex)
    {
      var handler = ActionFailed;
      if (handler != null)
        handler(_currentAction, ex);
      else
        throw new ApplicationException("ActionFailed event for ActionQueue not handled, and exception occurred. Queue stopped.", ex);
    }

    private void WaitForAction()
    {
      lock (_queueLockObject)
      {
        while (_queue.Count == 0)
          Monitor.Wait(_queueLockObject);

        _currentAction = _queue[0];
        _queue.RemoveAt(0);
      }
    }

    public void EnqueueAction(TQeueuedAction action)
    {
      if (_stopped)
        throw new ObjectDisposedException("ActionQueue");

      lock (_queueLockObject)
      {
        RemoveRedundantActions(action);
        _queue.Add(action);

        Monitor.PulseAll(_queueLockObject);
      }
    }

    private void RemoveRedundantActions(TQeueuedAction action)
    {
      for (int i = _queue.Count - 1; i >= 0; i--)
      {
        if (action.Replaces(_queue[i]))
          _queue.RemoveAt(i);
      }
    }

    public void Dispose()
    {
      _stopped = true;
    }

    public abstract class QueuedAction
    {
      public QueuedAction(Action action)
      {
        Action = action;
      }

      public Action Action { get; set; }

      public abstract bool Replaces(TQeueuedAction queuedAction);
    }

  }
}
